# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.contrib.piecewise.tests.common_tests as ct
from pyomo.contrib.piecewise.tests.models import make_log_x_model
from pyomo.contrib.piecewise.triangulations import Triangulation
from pyomo.core.base import TransformationFactory
from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.environ import (
    Constraint,
    SolverFactory,
    Var,
    ConcreteModel,
    Objective,
    log,
    value,
    minimize,
)
from pyomo.contrib.piecewise import PiecewiseLinearFunction
import itertools


class TestTransformPiecewiseModelToIncrementalMIP(unittest.TestCase):

    def check_pw_log(self, m):
        z = m.pw_log.get_transformation_var(m.log_expr)
        self.assertIsInstance(z, Var)
        log_block = z.parent_block()

        # Vars: three deltas, two y binaries, one substitute var
        self.assertEqual(len(log_block.component_map(Var)), 3)
        self.assertIsInstance(log_block.delta, Var)
        self.assertEqual(len(log_block.delta), 3)
        self.assertIsInstance(log_block.y_binaries, Var)
        self.assertEqual(len(log_block.y_binaries), 2)
        self.assertIsInstance(log_block.substitute_var, Var)
        self.assertEqual(len(log_block.substitute_var), 1)

        # Constraints: 2 delta below y, 2 y below delta, one each of the three others
        self.assertEqual(len(log_block.component_map(Constraint)), 5)
        self.assertIsInstance(log_block.deltas_below_y, Constraint)
        self.assertEqual(len(log_block.deltas_below_y), 2)
        self.assertIsInstance(log_block.y_below_delta, Constraint)
        self.assertEqual(len(log_block.y_below_delta), 2)
        self.assertIsInstance(log_block.delta_one_constraint, Constraint)
        self.assertEqual(len(log_block.delta_one_constraint), 1)
        self.assertIsInstance(log_block.x_constraint, Constraint)
        self.assertEqual(len(log_block.x_constraint), 1)
        self.assertIsInstance(log_block.set_substitute, Constraint)
        self.assertEqual(len(log_block.set_substitute), 1)

        assertExpressionsEqual(
            self,
            log_block.x_constraint[0].expr,
            m.x
            == 1
            + (
                log_block.delta[0, 1] * (3 - 1)
                + log_block.delta[1, 1] * (6 - 3)
                + log_block.delta[2, 1] * (10 - 6)
            ),
        )
        assertExpressionsEqual(
            self,
            log_block.set_substitute.expr,
            log_block.substitute_var
            == m.f1(1)
            + (
                log_block.delta[0, 1] * (m.f2(3) - m.f1(1))
                + log_block.delta[1, 1] * (m.f3(6) - m.f2(3))
                + log_block.delta[2, 1] * (m.f3(10) - m.f3(6))
            ),
            places=10,
        )
        assertExpressionsEqual(
            self, log_block.delta_one_constraint.expr, log_block.delta[0, 1] <= 1
        )
        assertExpressionsEqual(
            self,
            log_block.deltas_below_y[0].expr,
            log_block.delta[1, 1] <= log_block.y_binaries[0],
        )
        assertExpressionsEqual(
            self,
            log_block.deltas_below_y[1].expr,
            log_block.delta[2, 1] <= log_block.y_binaries[1],
        )
        assertExpressionsEqual(
            self,
            log_block.y_below_delta[0].expr,
            log_block.y_binaries[0] <= log_block.delta[0, 1],
        )
        assertExpressionsEqual(
            self,
            log_block.y_below_delta[1].expr,
            log_block.y_binaries[1] <= log_block.delta[1, 1],
        )

    def check_pw_paraboloid(self, m):
        z = m.pw_paraboloid.get_transformation_var(m.paraboloid_expr)
        self.assertIsInstance(z, Var)
        paraboloid_block = z.parent_block()

        # Vars: 8 deltas (2 per simplex), 3 y binaries, one substitute var
        self.assertEqual(len(paraboloid_block.component_map(Var)), 3)
        self.assertIsInstance(paraboloid_block.delta, Var)
        self.assertEqual(len(paraboloid_block.delta), 8)
        self.assertIsInstance(paraboloid_block.y_binaries, Var)
        self.assertEqual(len(paraboloid_block.y_binaries), 3)
        self.assertIsInstance(paraboloid_block.substitute_var, Var)
        self.assertEqual(len(paraboloid_block.substitute_var), 1)

        # Constraints: 3 delta below y, 3 y below delta, two x constraints (two
        # coordinates), one each of the three others
        self.assertEqual(len(paraboloid_block.component_map(Constraint)), 5)
        self.assertIsInstance(paraboloid_block.deltas_below_y, Constraint)
        self.assertEqual(len(paraboloid_block.deltas_below_y), 3)
        self.assertIsInstance(paraboloid_block.y_below_delta, Constraint)
        self.assertEqual(len(paraboloid_block.y_below_delta), 3)
        self.assertIsInstance(paraboloid_block.delta_one_constraint, Constraint)
        self.assertEqual(len(paraboloid_block.delta_one_constraint), 1)
        self.assertIsInstance(paraboloid_block.x_constraint, Constraint)
        self.assertEqual(len(paraboloid_block.x_constraint), 2)
        self.assertIsInstance(paraboloid_block.set_substitute, Constraint)
        self.assertEqual(len(paraboloid_block.set_substitute), 1)

    ordered_simplices = [
        [(0, 1), (3, 1), (3, 4)],
        [(3, 4), (0, 1), (0, 4)],
        [(0, 4), (0, 7), (3, 4)],
        [(3, 4), (3, 7), (0, 7)],
    ]

    # Test methods using the common_tests.py code.
    def test_transformation_do_not_descend(self):
        ct.check_transformation_do_not_descend(
            self,
            'contrib.piecewise.incremental',
            make_log_x_model(simplices=self.ordered_simplices),
        )

    def test_transformation_PiecewiseLinearFunction_targets(self):
        ct.check_transformation_PiecewiseLinearFunction_targets(
            self,
            'contrib.piecewise.incremental',
            make_log_x_model(simplices=self.ordered_simplices),
        )

    def test_descend_into_expressions(self):
        ct.check_descend_into_expressions(
            self,
            'contrib.piecewise.incremental',
            make_log_x_model(simplices=self.ordered_simplices),
        )

    def test_descend_into_expressions_constraint_target(self):
        ct.check_descend_into_expressions_constraint_target(
            self,
            'contrib.piecewise.incremental',
            make_log_x_model(simplices=self.ordered_simplices),
        )

    def test_descend_into_expressions_objective_target(self):
        ct.check_descend_into_expressions_objective_target(
            self,
            'contrib.piecewise.incremental',
            make_log_x_model(simplices=self.ordered_simplices),
        )

    @unittest.skipUnless(SolverFactory('gurobi').available(), 'Gurobi is not available')
    @unittest.skipUnless(SolverFactory('gurobi').license_is_valid(), 'No license')
    def test_solve_log_model(self):
        m = make_log_x_model(simplices=self.ordered_simplices)
        TransformationFactory('contrib.piecewise.incremental').apply_to(m)
        TransformationFactory('gdp.bigm').apply_to(m)
        SolverFactory('gurobi').solve(m)
        ct.check_log_x_model_soln(self, m)

    # Failed during development when ordered j1 vertex ordering got broken
    @unittest.skipUnless(SolverFactory('gurobi').available(), 'Gurobi is not available')
    @unittest.skipUnless(SolverFactory('gurobi').license_is_valid(), 'No license')
    def test_solve_product_model(self):
        m = ConcreteModel()
        m.x1 = Var(bounds=(0.5, 5))
        m.x2 = Var(bounds=(0.9, 0.95))
        pts = list(itertools.product([0.5, 2.75, 5], [0.9, 0.925, 0.95]))
        m.pwlf = PiecewiseLinearFunction(
            points=pts,
            function=lambda x, y: x * y,
            triangulation=Triangulation.OrderedJ1,
        )
        m.obj = Objective(sense=minimize, expr=m.pwlf(m.x1, m.x2))
        TransformationFactory("contrib.piecewise.incremental").apply_to(m)
        SolverFactory('gurobi').solve(m)
        self.assertAlmostEqual(0.45, value(m.obj))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for indexed components
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest
from io import StringIO

from pyomo.environ import (
    ConcreteModel,
    Block,
    Var,
    Set,
    RangeSet,
    Param,
    value,
    NonNegativeIntegers,
)
from pyomo.common.collections import ComponentSet
from pyomo.common.log import LoggingIntercept
from pyomo.core.base.var import IndexedVar
from pyomo.core.base.set import (
    SetProduct,
    FiniteSetOf,
    OrderedSetOf,
    UnknownSetDimen,
    normalize_index,
)
from pyomo.core.base.indexed_component import UnindexedComponent_set, IndexedComponent
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice
from pyomo.core.base.reference import (
    _ReferenceDict,
    _ReferenceSet,
    Reference,
    UnindexedComponent_ReferenceSet,
)


class TestReferenceDict(unittest.TestCase):
    def setUp(self):
        self.m = m = ConcreteModel()

        @m.Block([1, 2], [4, 5])
        def b(b, i, j):
            b.x = Var([7, 8], [10, 11], initialize=0)
            b.y = Var([7, 8], initialize=0)
            b.z = Var()

        @m.Block([1, 2])
        def c(b, i):
            b.x = Var([7, 8], [10, 11], initialize=0)
            b.y = Var([7, 8], initialize=0)
            b.z = Var()

    def _lookupTester(self, _slice, key, ans):
        rd = _ReferenceDict(_slice)
        self.assertIn(key, rd)
        self.assertIs(rd[key], ans)

        if len(key) == 1:
            self.assertIn(key[0], rd)
            self.assertIs(rd[key[0]], ans)

        self.assertNotIn(None, rd)
        with self.assertRaises(KeyError):
            rd[None]

        for i in range(len(key)):
            _ = tuple([0] * i)
            self.assertNotIn(_, rd)
            with self.assertRaises(KeyError):
                rd[_]

    def test_simple_lookup(self):
        m = self.m

        self._lookupTester(m.b[:, :].x[:, :], (1, 5, 7, 10), m.b[1, 5].x[7, 10])
        self._lookupTester(m.b[:, 4].x[8, :], (1, 10), m.b[1, 4].x[8, 10])
        self._lookupTester(m.b[:, 4].x[8, 10], (1,), m.b[1, 4].x[8, 10])
        self._lookupTester(m.b[1, 4].x[8, :], (10,), m.b[1, 4].x[8, 10])

        self._lookupTester(m.b[:, :].y[:], (1, 5, 7), m.b[1, 5].y[7])
        self._lookupTester(m.b[:, 4].y[:], (1, 7), m.b[1, 4].y[7])
        self._lookupTester(m.b[:, 4].y[8], (1,), m.b[1, 4].y[8])

        self._lookupTester(m.b[:, :].z, (1, 5), m.b[1, 5].z)
        self._lookupTester(m.b[:, 4].z, (1,), m.b[1, 4].z)

        self._lookupTester(m.c[:].x[:, :], (1, 7, 10), m.c[1].x[7, 10])
        self._lookupTester(m.c[:].x[8, :], (1, 10), m.c[1].x[8, 10])
        self._lookupTester(m.c[:].x[8, 10], (1,), m.c[1].x[8, 10])
        self._lookupTester(m.c[1].x[:, :], (8, 10), m.c[1].x[8, 10])
        self._lookupTester(m.c[1].x[8, :], (10,), m.c[1].x[8, 10])

        self._lookupTester(m.c[:].y[:], (1, 7), m.c[1].y[7])
        self._lookupTester(m.c[:].y[8], (1,), m.c[1].y[8])
        self._lookupTester(m.c[1].y[:], (8,), m.c[1].y[8])

        self._lookupTester(m.c[:].z, (1,), m.c[1].z)

        m.jagged_set = Set(initialize=[1, (2, 3)], dimen=None)
        m.jb = Block(m.jagged_set)
        m.jb[1].x = Var([1, 2, 3])
        m.jb[2, 3].x = Var([1, 2, 3])
        self._lookupTester(m.jb[...], (1,), m.jb[1])
        self._lookupTester(m.jb[...].x[:], (1, 2), m.jb[1].x[2])
        self._lookupTester(m.jb[...].x[:], (2, 3, 2), m.jb[2, 3].x[2])

        rd = _ReferenceDict(m.jb[:, :, :].x[:])
        with self.assertRaises(KeyError):
            rd[2, 3, 4, 2]
        rd = _ReferenceDict(m.b[:, 4].x[:])
        with self.assertRaises(KeyError):
            rd[1, 0]

    def test_len(self):
        m = self.m

        rd = _ReferenceDict(m.b[:, :].x[:, :])
        self.assertEqual(len(rd), 2 * 2 * 2 * 2)

        rd = _ReferenceDict(m.b[:, 4].x[8, :])
        self.assertEqual(len(rd), 2 * 2)

    def test_iterators(self):
        m = self.m
        rd = _ReferenceDict(m.b[:, 4].x[8, :])

        self.assertEqual(list(rd), [(1, 10), (1, 11), (2, 10), (2, 11)])
        self.assertEqual(list(rd.keys()), [(1, 10), (1, 11), (2, 10), (2, 11)])
        self.assertEqual(
            list(rd.values()),
            [
                m.b[1, 4].x[8, 10],
                m.b[1, 4].x[8, 11],
                m.b[2, 4].x[8, 10],
                m.b[2, 4].x[8, 11],
            ],
        )
        self.assertEqual(
            list(rd.items()),
            [
                ((1, 10), m.b[1, 4].x[8, 10]),
                ((1, 11), m.b[1, 4].x[8, 11]),
                ((2, 10), m.b[2, 4].x[8, 10]),
                ((2, 11), m.b[2, 4].x[8, 11]),
            ],
        )

    def test_ordered_iterators(self):
        # Test slice; common indexing set
        m = ConcreteModel()
        m.I = Set(initialize=[3, 2])
        m.b = Block([1, 0])
        m.b[1].x = Var(m.I)
        m.b[0].x = Var(m.I)
        m.y = Reference(m.b[:].x[:])
        self.assertEqual(list(m.y.index_set().subsets()), [m.b.index_set(), m.I])
        self.assertEqual(list(m.y), [(1, 3), (1, 2), (0, 3), (0, 2)])
        self.assertEqual(list(m.y.keys()), [(1, 3), (1, 2), (0, 3), (0, 2)])
        self.assertEqual(
            list(m.y.values()), [m.b[1].x[3], m.b[1].x[2], m.b[0].x[3], m.b[0].x[2]]
        )
        self.assertEqual(
            list(m.y.items()),
            [
                ((1, 3), m.b[1].x[3]),
                ((1, 2), m.b[1].x[2]),
                ((0, 3), m.b[0].x[3]),
                ((0, 2), m.b[0].x[2]),
            ],
        )
        self.assertEqual(list(m.y.keys(True)), [(0, 2), (0, 3), (1, 2), (1, 3)])
        self.assertEqual(
            list(m.y.values(True)), [m.b[0].x[2], m.b[0].x[3], m.b[1].x[2], m.b[1].x[3]]
        )
        self.assertEqual(
            list(m.y.items(True)),
            [
                ((0, 2), m.b[0].x[2]),
                ((0, 3), m.b[0].x[3]),
                ((1, 2), m.b[1].x[2]),
                ((1, 3), m.b[1].x[3]),
            ],
        )

        # Test slice; ReferenceSet indexing set
        m = ConcreteModel()
        m.b = Block([1, 0])
        m.b[1].x = Var([3, 2])
        m.b[0].x = Var([5, 4])
        m.y = Reference(m.b[:].x[:])
        self.assertIs(type(m.y.index_set()), FiniteSetOf)
        self.assertEqual(list(m.y), [(1, 3), (1, 2), (0, 5), (0, 4)])
        self.assertEqual(list(m.y.keys()), [(1, 3), (1, 2), (0, 5), (0, 4)])
        self.assertEqual(
            list(m.y.values()), [m.b[1].x[3], m.b[1].x[2], m.b[0].x[5], m.b[0].x[4]]
        )
        self.assertEqual(
            list(m.y.items()),
            [
                ((1, 3), m.b[1].x[3]),
                ((1, 2), m.b[1].x[2]),
                ((0, 5), m.b[0].x[5]),
                ((0, 4), m.b[0].x[4]),
            ],
        )
        self.assertEqual(list(m.y.keys(True)), [(0, 4), (0, 5), (1, 2), (1, 3)])
        self.assertEqual(
            list(m.y.values(True)), [m.b[0].x[4], m.b[0].x[5], m.b[1].x[2], m.b[1].x[3]]
        )
        self.assertEqual(
            list(m.y.items(True)),
            [
                ((0, 4), m.b[0].x[4]),
                ((0, 5), m.b[0].x[5]),
                ((1, 2), m.b[1].x[2]),
                ((1, 3), m.b[1].x[3]),
            ],
        )

        # Test dict, ReferenceSet indexing set
        m = ConcreteModel()
        m.b = Block([1, 0])
        m.b[1].x = Var([3, 2])
        m.b[0].x = Var([5, 4])
        m.y = Reference(
            {
                (1, 3): m.b[1].x[3],
                (0, 5): m.b[0].x[5],
                (1, 2): m.b[1].x[2],
                (0, 4): m.b[0].x[4],
            }
        )
        self.assertIs(type(m.y.index_set()), FiniteSetOf)
        self.assertEqual(list(m.y), [(1, 3), (0, 5), (1, 2), (0, 4)])
        self.assertEqual(list(m.y.keys()), [(1, 3), (0, 5), (1, 2), (0, 4)])
        self.assertEqual(
            list(m.y.values()), [m.b[1].x[3], m.b[0].x[5], m.b[1].x[2], m.b[0].x[4]]
        )
        self.assertEqual(
            list(m.y.items()),
            [
                ((1, 3), m.b[1].x[3]),
                ((0, 5), m.b[0].x[5]),
                ((1, 2), m.b[1].x[2]),
                ((0, 4), m.b[0].x[4]),
            ],
        )
        self.assertEqual(list(m.y.keys(True)), [(0, 4), (0, 5), (1, 2), (1, 3)])
        self.assertEqual(
            list(m.y.values(True)), [m.b[0].x[4], m.b[0].x[5], m.b[1].x[2], m.b[1].x[3]]
        )
        self.assertEqual(
            list(m.y.items(True)),
            [
                ((0, 4), m.b[0].x[4]),
                ((0, 5), m.b[0].x[5]),
                ((1, 2), m.b[1].x[2]),
                ((1, 3), m.b[1].x[3]),
            ],
        )

    def test_nested_assignment(self):
        m = self.m

        rd = _ReferenceDict(m.b[:, :].x[:, :])
        self.assertEqual(sum(x.value for x in rd.values()), 0)
        rd[1, 5, 7, 10] = 10
        self.assertEqual(m.b[1, 5].x[7, 10].value, 10)
        self.assertEqual(sum(x.value for x in rd.values()), 10)

        rd = _ReferenceDict(m.b[:, 4].x[8, :])
        self.assertEqual(sum(x.value for x in rd.values()), 0)
        rd[1, 10] = 20
        self.assertEqual(m.b[1, 4].x[8, 10].value, 20)
        self.assertEqual(sum(x.value for x in rd.values()), 20)

    def test_attribute_assignment(self):
        m = self.m

        rd = _ReferenceDict(m.b[:, :].x[:, :].value)
        self.assertEqual(sum(x for x in rd.values()), 0)
        rd[1, 5, 7, 10] = 10
        self.assertEqual(m.b[1, 5].x[7, 10].value, 10)
        self.assertEqual(sum(x for x in rd.values()), 10)

        rd = _ReferenceDict(m.b[:, 4].x[8, :].value)
        self.assertEqual(sum(x for x in rd.values()), 0)
        rd[1, 10] = 20
        self.assertEqual(m.b[1, 4].x[8, 10].value, 20)
        self.assertEqual(sum(x for x in rd.values()), 20)

        m.x = Var([1, 2], initialize=0)
        rd = _ReferenceDict(m.x[:])
        self.assertEqual(sum(x.value for x in rd.values()), 0)
        rd[2] = 10
        self.assertEqual(m.x[1].value, 0)
        self.assertEqual(m.x[2].value, 10)
        self.assertEqual(sum(x.value for x in rd.values()), 10)

    def test_single_attribute_assignment(self):
        m = self.m

        rd = _ReferenceDict(m.b[1, 5].x[:, :])
        self.assertEqual(sum(x.value for x in rd.values()), 0)
        rd[7, 10].value = 10
        self.assertEqual(m.b[1, 5].x[7, 10].value, 10)
        self.assertEqual(sum(x.value for x in rd.values()), 10)

        rd = _ReferenceDict(m.b[1, 4].x[8, :])
        self.assertEqual(sum(x.value for x in rd.values()), 0)
        rd[10].value = 20
        self.assertEqual(m.b[1, 4].x[8, 10].value, 20)
        self.assertEqual(sum(x.value for x in rd.values()), 20)

    def test_nested_attribute_assignment(self):
        m = self.m

        rd = _ReferenceDict(m.b[:, :].x[:, :])
        self.assertEqual(sum(x.value for x in rd.values()), 0)
        rd[1, 5, 7, 10].value = 10
        self.assertEqual(m.b[1, 5].x[7, 10].value, 10)
        self.assertEqual(sum(x.value for x in rd.values()), 10)

        rd = _ReferenceDict(m.b[:, 4].x[8, :])
        self.assertEqual(sum(x.value for x in rd.values()), 0)
        rd[1, 10].value = 20
        self.assertEqual(m.b[1, 4].x[8, 10].value, 20)
        self.assertEqual(sum(x.value for x in rd.values()), 20)

    def test_single_deletion(self):
        m = self.m

        rd = _ReferenceDict(m.b[1, 5].x[:, :])
        self.assertEqual(len(list(x.value for x in rd.values())), 2 * 2)
        self.assertTrue((7, 10) in rd)
        del rd[7, 10]
        self.assertFalse((7, 10) in rd)
        self.assertEqual(len(list(x.value for x in rd.values())), 3)

        rd = _ReferenceDict(m.b[1, 4].x[8, :])
        self.assertEqual(len(list(x.value for x in rd.values())), 2)
        self.assertTrue((10) in rd)
        del rd[10]
        self.assertFalse(10 in rd)
        self.assertEqual(len(list(x.value for x in rd.values())), 2 - 1)

        with self.assertRaisesRegex(
            KeyError, r"\(8, 10\) is not valid for indexed component 'b\[1,4\].x'"
        ):
            del rd[10]

        rd = _ReferenceDict(m.b[1, :].x[8, 0])
        with self.assertRaisesRegex(
            KeyError, r"'\(8, 0\)' is not valid for indexed component 'b\[1,4\].x'"
        ):
            del rd[4]

    def test_nested_deletion(self):
        m = self.m

        rd = _ReferenceDict(m.b[:, :].x[:, :])
        self.assertEqual(len(list(x.value for x in rd.values())), 2 * 2 * 2 * 2)
        self.assertTrue((1, 5, 7, 10) in rd)
        del rd[1, 5, 7, 10]
        self.assertFalse((1, 5, 7, 10) in rd)
        self.assertEqual(len(list(x.value for x in rd.values())), 2 * 2 * 2 * 2 - 1)

        rd = _ReferenceDict(m.b[:, 4].x[8, :])
        self.assertEqual(len(list(x.value for x in rd.values())), 2 * 2)
        self.assertTrue((1, 10) in rd)
        del rd[1, 10]
        self.assertFalse((1, 10) in rd)
        self.assertEqual(len(list(x.value for x in rd.values())), 2 * 2 - 1)

    def test_attribute_deletion(self):
        m = self.m

        rd = _ReferenceDict(m.b[:, :].z)
        rd._slice.attribute_errors_generate_exceptions = False
        self.assertEqual(len(list(x.value for x in rd.values())), 2 * 2)
        self.assertTrue((1, 5) in rd)
        self.assertTrue(hasattr(m.b[1, 5], 'z'))
        self.assertTrue(hasattr(m.b[2, 5], 'z'))
        del rd[1, 5]
        self.assertFalse((1, 5) in rd)
        self.assertFalse(hasattr(m.b[1, 5], 'z'))
        self.assertTrue(hasattr(m.b[2, 5], 'z'))
        self.assertEqual(len(list(x.value for x in rd.values())), 3)

        rd = _ReferenceDict(m.b[2, :].z)
        rd._slice.attribute_errors_generate_exceptions = False
        self.assertEqual(len(list(x.value for x in rd.values())), 2)
        self.assertTrue(5 in rd)
        self.assertTrue(hasattr(m.b[2, 4], 'z'))
        self.assertTrue(hasattr(m.b[2, 5], 'z'))
        del rd[5]
        self.assertFalse(5 in rd)
        self.assertTrue(hasattr(m.b[2, 4], 'z'))
        self.assertFalse(hasattr(m.b[2, 5], 'z'))
        self.assertEqual(len(list(x.value for x in rd.values())), 2 - 1)

    def test_deprecations(self):
        m = self.m
        rd = _ReferenceDict(m.b[:, :].z)

        items = rd.items()
        with LoggingIntercept() as LOG:
            iteritems = rd.iteritems()
        self.assertIs(type(items), type(iteritems))
        self.assertEqual(list(items), list(iteritems))
        self.assertIn(
            "DEPRECATED: The iteritems method is deprecated. Use dict.items",
            LOG.getvalue(),
        )

        values = rd.values()
        with LoggingIntercept() as LOG:
            itervalues = rd.itervalues()
        self.assertIs(type(values), type(itervalues))
        self.assertEqual(list(values), list(itervalues))
        self.assertIn(
            "DEPRECATED: The itervalues method is deprecated. Use dict.values",
            LOG.getvalue(),
        )


class TestReferenceSet(unittest.TestCase):
    def test_str(self):
        m = ConcreteModel()

        @m.Block([1, 2], [4, 5])
        def b(b, i, j):
            b.x = Var([7, 8], [10, 11], initialize=0)
            b.y = Var([7, 8], initialize=0)
            b.z = Var()

        rs = _ReferenceSet(m.b[:, 5].z)
        self.assertEqual(str(rs), 'ReferenceSet(b[:, 5].z)')

    def test_lookup_and_iter_dense_data(self):
        m = ConcreteModel()

        @m.Block([1, 2], [4, 5])
        def b(b, i, j):
            b.x = Var([7, 8], [10, 11], initialize=0)
            b.y = Var([7, 8], initialize=0)
            b.z = Var()

        rs = _ReferenceSet(m.b[:, 5].z)
        self.assertNotIn((0,), rs)
        self.assertIn(1, rs)
        self.assertIn((1,), rs)
        self.assertEqual(len(rs), 2)
        self.assertEqual(list(rs), [1, 2])

        rs = _ReferenceSet(m.b[:, 5].bad)
        self.assertNotIn((0,), rs)
        self.assertNotIn((1,), rs)
        self.assertEqual(len(rs), 0)
        self.assertEqual(list(rs), [])

        @m.Block([1, 2, 3])
        def d(b, i):
            if i % 2:
                b.x = Var(range(i))

        rs = _ReferenceSet(m.d[:].x[:])
        self.assertIn((1, 0), rs)
        self.assertIn((3, 0), rs)
        self.assertNotIn((2, 0), rs)
        self.assertEqual(len(rs), 4)
        self.assertEqual(list(rs), [(1, 0), (3, 0), (3, 1), (3, 2)])

        rs = _ReferenceSet(m.d[...].x[...])
        self.assertIn((1, 0), rs)
        self.assertIn((3, 0), rs)
        self.assertNotIn((2, 0), rs)
        self.assertEqual(len(rs), 4)
        self.assertEqual(list(rs), [(1, 0), (3, 0), (3, 1), (3, 2)])

        # Test the SliceEllipsisError case (lookup into a jagged set
        # with an ellipsis)

        m.e_index = Set(initialize=[2, (2, 3)], dimen=None)

        @m.Block(m.e_index)
        def e(b, *args):
            b.x_index = Set(initialize=[1, (3, 4)], dimen=None)
            b.x = Var(b.x_index)

        rs = _ReferenceSet(m.e[...].x[...])
        self.assertIn((2, 1), rs)
        self.assertIn((2, 3, 1), rs)
        self.assertIn((2, 3, 4), rs)
        self.assertNotIn((2, 3, 5), rs)
        self.assertEqual(len(rs), 4)
        self.assertEqual(list(rs), [(2, 1), (2, 3, 4), (2, 3, 1), (2, 3, 3, 4)])

        # Make sure scalars and tuples work for jagged sets
        rs = _ReferenceSet(m.e[...])
        self.assertIn(2, rs)
        self.assertIn((2,), rs)
        self.assertNotIn(0, rs)
        self.assertNotIn((0,), rs)

    def test_lookup_and_iter_sparse_data(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.x = Var(m.I, m.I, dense=False)

        rd = _ReferenceDict(m.x[...])
        rs = _ReferenceSet(m.x[...])
        self.assertEqual(len(rd), 0)
        # Note: we will periodically re-check the dict to ensure
        # iteration doesn't accidentally declare data
        self.assertEqual(len(rd), 0)

        self.assertEqual(len(rs), 9)
        self.assertEqual(len(rd), 0)

        self.assertIn((1, 1), rs)
        self.assertEqual(len(rd), 0)
        self.assertEqual(len(rs), 9)

    def test_otdered_sorted_iter(self):
        # Test ordered reference
        m = ConcreteModel()

        @m.Block([2, 1], [4, 5])
        def b(b, i, j):
            b.x = Var([8, 7], initialize=0)

        rs = _ReferenceSet(m.b[...].x[:])
        self.assertEqual(
            list(rs),
            [
                (2, 4, 8),
                (2, 4, 7),
                (2, 5, 8),
                (2, 5, 7),
                (1, 4, 8),
                (1, 4, 7),
                (1, 5, 8),
                (1, 5, 7),
            ],
        )

        rs = _ReferenceSet(m.b[...].x[:])
        self.assertEqual(
            list(rs.ordered_iter()),
            [
                (2, 4, 8),
                (2, 4, 7),
                (2, 5, 8),
                (2, 5, 7),
                (1, 4, 8),
                (1, 4, 7),
                (1, 5, 8),
                (1, 5, 7),
            ],
        )

        rs = _ReferenceSet(m.b[...].x[:])
        self.assertEqual(
            list(rs.sorted_iter()),
            [
                (1, 4, 7),
                (1, 4, 8),
                (1, 5, 7),
                (1, 5, 8),
                (2, 4, 7),
                (2, 4, 8),
                (2, 5, 7),
                (2, 5, 8),
            ],
        )

        # Test unordered reference
        m = ConcreteModel()
        m.I = FiniteSetOf([2, 1])
        m.J = FiniteSetOf([4, 5])
        m.K = FiniteSetOf([8, 7])

        @m.Block(m.I, m.J)
        def b(b, i, j):
            b.x = Var(m.K, initialize=0)

        rs = _ReferenceSet(m.b[...].x[:])
        self.assertEqual(
            list(rs),
            [
                (2, 4, 8),
                (2, 4, 7),
                (2, 5, 8),
                (2, 5, 7),
                (1, 4, 8),
                (1, 4, 7),
                (1, 5, 8),
                (1, 5, 7),
            ],
        )

        rs = _ReferenceSet(m.b[...].x[:])
        self.assertEqual(
            list(rs.ordered_iter()),
            [
                (1, 4, 7),
                (1, 4, 8),
                (1, 5, 7),
                (1, 5, 8),
                (2, 4, 7),
                (2, 4, 8),
                (2, 5, 7),
                (2, 5, 8),
            ],
        )

        rs = _ReferenceSet(m.b[...].x[:])
        self.assertEqual(
            list(rs.sorted_iter()),
            [
                (1, 4, 7),
                (1, 4, 8),
                (1, 5, 7),
                (1, 5, 8),
                (2, 4, 7),
                (2, 4, 8),
                (2, 5, 7),
                (2, 5, 8),
            ],
        )


class TestReference(unittest.TestCase):
    def test_constructor_error(self):
        m = ConcreteModel()
        m.x = Var([1, 2])

        class Foo:
            pass

        self.assertRaisesRegex(
            TypeError,
            "First argument to Reference constructors must be a "
            r"component, component slice, Sequence, or Mapping \(received Foo",
            Reference,
            Foo(),
        )
        self.assertRaisesRegex(
            TypeError,
            "First argument to Reference constructors must be a "
            r"component, component slice, Sequence, or Mapping \(received int",
            Reference,
            5,
        )
        self.assertRaisesRegex(
            TypeError,
            "First argument to Reference constructors must be a "
            r"component, component slice, Sequence, or Mapping \(received None",
            Reference,
            None,
        )

    def test_component_reference(self):
        m = ConcreteModel()
        m.x = Var()
        m.r = Reference(m.x)

        self.assertIs(m.r.ctype, Var)
        self.assertIsNot(m.r.index_set(), m.x.index_set())
        self.assertIs(m.x.index_set(), UnindexedComponent_set)
        self.assertIs(m.r.index_set(), UnindexedComponent_ReferenceSet)
        self.assertEqual(len(m.r), 1)
        self.assertTrue(m.r.is_indexed())
        self.assertIn(None, m.r)
        self.assertNotIn(1, m.r)
        self.assertIs(m.r[None], m.x)
        with self.assertRaises(KeyError):
            m.r[1]

        m.s = Reference(m.x[:])

        self.assertIs(m.s.ctype, Var)
        self.assertIsNot(m.s.index_set(), m.x.index_set())
        self.assertIs(m.x.index_set(), UnindexedComponent_set)
        self.assertIs(type(m.s.index_set()), OrderedSetOf)
        self.assertEqual(len(m.s), 1)
        self.assertTrue(m.s.is_indexed())
        self.assertIn(None, m.s)
        self.assertNotIn(1, m.s)
        self.assertIs(m.s[None], m.x)
        with self.assertRaises(KeyError):
            m.s[1]

        m.y = Var([1, 2])
        m.t = Reference(m.y)

        self.assertIs(m.t.ctype, Var)
        self.assertIs(m.t.index_set(), m.y.index_set())
        self.assertEqual(len(m.t), 2)
        self.assertTrue(m.t.is_indexed())
        self.assertNotIn(None, m.t)
        self.assertIn(1, m.t)
        self.assertIn(2, m.t)
        self.assertIs(m.t[1], m.y[1])
        with self.assertRaises(KeyError):
            m.t[3]

    def test_component_data_reference(self):
        m = ConcreteModel()
        m.y = Var([1, 2])
        m.r = Reference(m.y[2])

        self.assertIs(m.r.ctype, Var)
        self.assertIsNot(m.r.index_set(), m.y.index_set())
        self.assertIs(m.r.index_set(), UnindexedComponent_ReferenceSet)
        self.assertEqual(len(m.r), 1)
        self.assertTrue(m.r.is_reference())
        self.assertTrue(m.r.is_indexed())
        self.assertIn(None, m.r)
        self.assertNotIn(1, m.r)
        self.assertIs(m.r[None], m.y[2])
        with self.assertRaises(KeyError):
            m.r[2]

    def test_component_data_reference_clone(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.x = Var([1, 2])
        m.c = Block()
        m.c.r1 = Reference(m.b.x[2])
        m.c.r2 = Reference(m.b.x)

        self.assertIs(m.c.r1[None], m.b.x[2])
        m.d = m.c.clone()
        self.assertIs(m.d.r1[None], m.b.x[2])
        self.assertIs(m.d.r2[1], m.b.x[1])
        self.assertIs(m.d.r2[2], m.b.x[2])

        i = m.clone()
        self.assertIs(i.c.r1[None], i.b.x[2])
        self.assertIs(i.c.r2[1], i.b.x[1])
        self.assertIs(i.c.r2[2], i.b.x[2])
        self.assertIsNot(i.c.r1[None], m.b.x[2])
        self.assertIsNot(i.c.r2[1], m.b.x[1])
        self.assertIsNot(i.c.r2[2], m.b.x[2])
        self.assertIs(i.d.r1[None], i.b.x[2])
        self.assertIs(i.d.r2[1], i.b.x[1])
        self.assertIs(i.d.r2[2], i.b.x[2])

    def test_reference_var_pprint(self):
        m = ConcreteModel()
        m.x = Var([1, 2], initialize={1: 4, 2: 8})
        m.r = Reference(m.x)
        buf = StringIO()
        m.r.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue(),
            """r : Size=2, Index={1, 2}, ReferenceTo=x
    Key : Lower : Value : Upper : Fixed : Stale : Domain
      1 :  None :     4 :  None : False : False :  Reals
      2 :  None :     8 :  None : False : False :  Reals
""",
        )
        m.s = Reference(m.x[:, ...])
        buf = StringIO()
        m.s.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue(),
            """s : Size=2, Index={1, 2}, ReferenceTo=x[:, ...]
    Key : Lower : Value : Upper : Fixed : Stale : Domain
      1 :  None :     4 :  None : False : False :  Reals
      2 :  None :     8 :  None : False : False :  Reals
""",
        )

    def test_reference_indexedcomponent_pprint(self):
        m = ConcreteModel()
        m.x = Var([1, 2], initialize={1: 4, 2: 8})
        m.r = Reference(m.x, ctype=IndexedComponent)
        buf = StringIO()
        m.r.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue(),
            """r : Size=2, Index={1, 2}, ReferenceTo=x
    Key : Object
      1 : <class 'pyomo.core.base.var.VarData'>
      2 : <class 'pyomo.core.base.var.VarData'>
""",
        )
        m.s = Reference(m.x[:, ...], ctype=IndexedComponent)
        buf = StringIO()
        m.s.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue(),
            """s : Size=2, Index={1, 2}, ReferenceTo=x[:, ...]
    Key : Object
      1 : <class 'pyomo.core.base.var.VarData'>
      2 : <class 'pyomo.core.base.var.VarData'>
""",
        )

    def test_single_reference(self):
        m = ConcreteModel()
        m.b = Block([1, 2])
        m.b[1].x = Var(bounds=(1, None))
        m.b[2].x = Var(bounds=(2, None))
        m.r = Reference(m.b[:].x)

        self.assertIs(m.r.ctype, Var)
        self.assertIs(m.r.index_set(), m.b.index_set())
        self.assertEqual(len(m.r), 2)
        self.assertEqual(m.r[1].lb, 1)
        self.assertEqual(m.r[2].lb, 2)
        self.assertIn(1, m.r)
        self.assertIn(2, m.r)
        self.assertNotIn(3, m.r)
        with self.assertRaises(KeyError):
            m.r[3]

    def test_nested_reference(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 2])
        m.J = Set(initialize=[3, 4])

        @m.Block(m.I)
        def b(b, i):
            b.x = Var(b.model().J, bounds=(i, None))

        m.r = Reference(m.b[:].x[:])

        self.assertIs(m.r.ctype, Var)
        self.assertIsInstance(m.r.index_set(), SetProduct)
        self.assertIs(m.r.index_set().set_tuple[0], m.I)
        self.assertIs(m.r.index_set().set_tuple[1], m.J)
        self.assertEqual(len(m.r), 2 * 2)
        self.assertEqual(m.r[1, 3].lb, 1)
        self.assertEqual(m.r[2, 4].lb, 2)
        self.assertIn((1, 3), m.r)
        self.assertIn((2, 4), m.r)
        self.assertNotIn(0, m.r)
        self.assertNotIn((1, 0), m.r)
        self.assertNotIn((1, 3, 0), m.r)
        with self.assertRaises(KeyError):
            m.r[0]

    def test_nested_reference_multidim_set(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 2])
        m.J = Set(initialize=[(3, 3), (4, 4)])

        @m.Block(m.I)
        def b(b, i):
            b.x = Var(b.model().J, bounds=(i, None))

        m.r = Reference(m.b[:].x[:, :])

        self.assertIs(m.r.ctype, Var)
        self.assertIsInstance(m.r.index_set(), SetProduct)
        self.assertIs(m.r.index_set().set_tuple[0], m.I)
        self.assertIs(m.r.index_set().set_tuple[1], m.J)
        self.assertEqual(len(m.r), 2 * 2)
        self.assertEqual(m.r[1, 3, 3].lb, 1)
        self.assertEqual(m.r[2, 4, 4].lb, 2)
        self.assertIn((1, 3, 3), m.r)
        self.assertIn((2, 4, 4), m.r)
        self.assertNotIn(0, m.r)
        self.assertNotIn((1, 0), m.r)
        self.assertNotIn((1, 3, 0), m.r)
        self.assertNotIn((1, 3, 3, 0), m.r)
        with self.assertRaises(KeyError):
            m.r[0]

    def test_nested_reference_partial_multidim_set(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 2])
        m.J = Set(initialize=[(3, 3), (4, 4)])

        @m.Block(m.I)
        def b(b, i):
            b.x = Var(b.model().J, bounds=(i, None))

        m.r = Reference(m.b[:].x[3, :])

        self.assertIs(m.r.ctype, Var)
        self.assertIs(type(m.r.index_set()), FiniteSetOf)
        self.assertEqual(len(m.r), 2 * 1)
        self.assertEqual(m.r[1, 3].lb, 1)
        self.assertEqual(m.r[2, 3].lb, 2)
        self.assertIn((1, 3), m.r)
        self.assertNotIn((2, 4), m.r)
        self.assertNotIn(0, m.r)
        self.assertNotIn((1, 0), m.r)
        self.assertNotIn((1, 3, 0), m.r)
        with self.assertRaises(KeyError):
            m.r[0]

    def test_nested_reference_nonuniform_indexes(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 2])
        m.J = Set(initialize=[3, 4])

        @m.Block(m.I)
        def b(b, i):
            b.x = Var([3, 4], bounds=(i, None))

        m.r = Reference(m.b[:].x[:])

        self.assertIs(m.r.ctype, Var)
        self.assertIs(type(m.r.index_set()), FiniteSetOf)
        self.assertEqual(len(m.r), 2 * 2)
        self.assertEqual(m.r[1, 3].lb, 1)
        self.assertEqual(m.r[2, 4].lb, 2)
        self.assertIn((1, 3), m.r)
        self.assertIn((2, 4), m.r)
        self.assertNotIn(0, m.r)
        self.assertNotIn((1, 0), m.r)
        self.assertNotIn((1, 3, 0), m.r)
        with self.assertRaises(KeyError):
            m.r[0]

    def test_nested_reference_nondimen_set(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 2])
        m.J = Set(initialize=[3, 4], dimen=None)

        @m.Block(m.I)
        def b(b, i):
            b.x = Var(b.model().J, bounds=(i, None))

        m.r = Reference(m.b[:].x[:])

        self.assertIs(m.r.ctype, Var)
        self.assertIs(type(m.r.index_set()), FiniteSetOf)
        self.assertEqual(len(m.r), 2 * 2)
        self.assertEqual(m.r[1, 3].lb, 1)
        self.assertEqual(m.r[2, 4].lb, 2)
        self.assertIn((1, 3), m.r)
        self.assertIn((2, 4), m.r)
        self.assertNotIn(0, m.r)
        self.assertNotIn((1, 0), m.r)
        self.assertNotIn((1, 3, 0), m.r)
        with self.assertRaises(KeyError):
            m.r[0]

    def test_nested_reference_nonuniform_index_size(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 2])
        m.J = Set(initialize=[3, 4])
        m.b = Block(m.I)
        m.b[1].x = Var([(3, 3), (3, 4), (4, 3), (4, 4)], bounds=(1, None))
        m.b[2].x = Var(m.J, m.J, bounds=(2, None))

        m.r = Reference(m.b[:].x[:, :])

        self.assertIs(m.r.ctype, Var)
        self.assertIs(type(m.r.index_set()), FiniteSetOf)
        self.assertEqual(len(m.r), 2 * 2 * 2)
        self.assertEqual(m.r[1, 3, 3].lb, 1)
        self.assertEqual(m.r[2, 4, 3].lb, 2)
        self.assertIn((1, 3, 3), m.r)
        self.assertIn((2, 4, 4), m.r)
        self.assertNotIn(0, m.r)
        self.assertNotIn((1, 0), m.r)
        self.assertNotIn((1, 3, 0), m.r)
        self.assertNotIn((1, 3, 3, 0), m.r)
        with self.assertRaises(KeyError):
            m.r[0]

    def test_nested_scalars(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.x = Var()
        m.r = Reference(m.b[:].x[:])
        self.assertEqual(len(m.r), 1)
        self.assertEqual(m.r.index_set().dimen, 2)
        base_sets = list(m.r.index_set().subsets())
        self.assertEqual(len(base_sets), 2)
        self.assertIs(type(base_sets[0]), OrderedSetOf)
        self.assertIs(type(base_sets[1]), OrderedSetOf)

    def test_ctype_detection(self):
        m = ConcreteModel()
        m.js = Set(initialize=[1, (2, 3)], dimen=None)
        m.b = Block([1, 2])
        m.b[1].x = Var(m.js)
        m.b[1].y = Var()
        m.b[1].z = Var([1, 2])
        m.b[2].x = Param(initialize=0)
        m.b[2].y = Var()
        m.b[2].z = Var([1, 2])

        m.x = Reference(m.b[:].x[...])
        self.assertIs(type(m.x), IndexedComponent)

        m.y = Reference(m.b[:].y[...])
        self.assertIs(type(m.y), IndexedVar)
        self.assertIs(m.y.ctype, Var)
        m.y1 = Reference(m.b[:].y[...], ctype=None)
        self.assertIs(type(m.y1), IndexedComponent)
        self.assertIs(m.y1.ctype, IndexedComponent)

        m.z = Reference(m.b[:].z)
        self.assertIs(type(m.z), IndexedComponent)
        self.assertIs(m.z.ctype, IndexedComponent)

    def test_reference_to_sparse(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.x = Var(m.I, m.I, dense=False)
        m.xx = Reference(m.x[...], ctype=Var)

        self.assertEqual(len(m.x), 0)
        self.assertNotIn((1, 1), m.x)
        self.assertNotIn((1, 1), m.xx)
        self.assertIn((1, 1), m.x.index_set())
        self.assertIn((1, 1), m.xx.index_set())
        self.assertEqual(len(m.x), 0)

        m.xx[1, 2]
        self.assertEqual(len(m.x), 1)
        self.assertIs(m.xx[1, 2], m.x[1, 2])
        self.assertEqual(len(m.x), 1)

        m.xx[1, 3] = 5
        self.assertEqual(len(m.x), 2)
        self.assertIs(m.xx[1, 3], m.x[1, 3])
        self.assertEqual(len(m.x), 2)
        self.assertEqual(value(m.x[1, 3]), 5)

        m.xx.add((1, 1))
        self.assertEqual(len(m.x), 3)
        self.assertIs(m.xx[1, 1], m.x[1, 1])
        self.assertEqual(len(m.x), 3)

    def test_nested_reference_to_sparse(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1])

        @m.Block(m.I)
        def b(b, i):
            b.x = Var(b.model().I, dense=False)

        m.xx = Reference(m.b[:].x[:], ctype=Var)
        m.I.add(2)
        m.I.add(3)

        self.assertEqual(len(m.b), 1)
        self.assertEqual(len(m.b[1].x), 0)
        self.assertIn(1, m.b)
        self.assertNotIn((1, 1), m.xx)
        self.assertIn(1, m.b[1].x.index_set())
        self.assertIn((1, 1), m.xx.index_set())
        self.assertEqual(len(m.b), 1)
        self.assertEqual(len(m.b[1].x), 0)

        m.xx[1, 2]
        self.assertEqual(len(m.b), 1)
        self.assertEqual(len(m.b[1].x), 1)
        self.assertIs(m.xx[1, 2], m.b[1].x[2])
        self.assertEqual(len(m.b), 1)
        self.assertEqual(len(m.b[1].x), 1)

        m.xx[1, 3] = 5
        self.assertEqual(len(m.b), 1)
        self.assertEqual(len(m.b[1].x), 2)
        self.assertIs(m.xx[1, 3], m.b[1].x[3])
        self.assertEqual(value(m.b[1].x[3]), 5)
        self.assertEqual(len(m.b), 1)
        self.assertEqual(len(m.b[1].x), 2)

        m.xx.add((1, 1))
        self.assertEqual(len(m.b), 1)
        self.assertEqual(len(m.b[1].x), 3)
        self.assertIs(m.xx[1, 1], m.b[1].x[1])
        self.assertEqual(len(m.b), 1)
        self.assertEqual(len(m.b[1].x), 3)

        # While (2,2) appears to be a valid member of the slice, because
        # 2 was not in the Set when the Block rule fired, there is no
        # m.b[2] block data.  Accessing m.xx[2,1] will construct the
        # b[2] block data, fire the rule, and then add the new value to
        # the Var x.
        self.assertEqual(len(m.xx), 3)
        m.xx[2, 2] = 10
        self.assertEqual(len(m.b), 2)
        self.assertEqual(len(list(m.b[2].component_objects())), 1)
        self.assertEqual(len(m.xx), 4)
        self.assertIs(m.xx[2, 2], m.b[2].x[2])
        self.assertEqual(value(m.b[2].x[2]), 10)

    def test_insert_var(self):
        m = ConcreteModel()
        m.T = Set(initialize=[1, 5])
        m.x = Var(m.T, initialize=lambda m, i: i)

        @m.Block(m.T)
        def b(b, i):
            b.y = Var(initialize=lambda b: 10 * b.index())

        ref_x = Reference(m.x[:])
        ref_y = Reference(m.b[:].y)

        self.assertEqual(len(m.x), 2)
        self.assertEqual(len(ref_x), 2)
        self.assertEqual(len(m.b), 2)
        self.assertEqual(len(ref_y), 2)
        self.assertEqual(value(ref_x[1]), 1)
        self.assertEqual(value(ref_x[5]), 5)
        self.assertEqual(value(ref_y[1]), 10)
        self.assertEqual(value(ref_y[5]), 50)

        m.T.add(2)
        _x = ref_x[2]
        self.assertEqual(len(m.x), 3)
        self.assertIs(_x, m.x[2])
        self.assertEqual(value(_x), 2)
        self.assertEqual(value(m.x[2]), 2)
        self.assertEqual(value(ref_x[2]), 2)

        _y = ref_y[2]
        self.assertEqual(len(m.b), 3)
        self.assertIs(_y, m.b[2].y)
        self.assertEqual(value(_y), 20)
        self.assertEqual(value(ref_y[2]), 20)
        self.assertEqual(value(m.b[2].y), 20)

    def test_reference_to_dict(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([1, 2, 3])
        m.r = Reference({1: m.x, 'a': m.y[2], 3: m.y[1]})
        self.assertFalse(m.r.index_set().isordered())
        self.assertEqual(len(m.r), 3)
        self.assertEqual(set(m.r.keys()), {1, 3, 'a'})
        self.assertEqual(
            ComponentSet(m.r.values()), ComponentSet([m.x, m.y[2], m.y[1]])
        )
        # You can delete something from the reference
        del m.r[1]
        self.assertEqual(len(m.r), 2)
        self.assertEqual(set(m.r.keys()), {3, 'a'})
        self.assertEqual(ComponentSet(m.r.values()), ComponentSet([m.y[2], m.y[1]]))
        # But not add it back
        with self.assertRaisesRegex(
            KeyError, "Index '1' is not valid for indexed component 'r'"
        ):
            m.r[1] = m.x

    def test_reference_to_list(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([1, 2, 3])
        m.r = Reference([m.x, m.y[2], m.y[1]])
        self.assertTrue(m.r.index_set().isordered())
        self.assertEqual(len(m.r), 3)
        self.assertEqual(list(m.r.keys()), [0, 1, 2])
        self.assertEqual(list(m.r.values()), [m.x, m.y[2], m.y[1]])
        # You can delete something from the reference
        del m.r[1]
        self.assertEqual(len(m.r), 2)
        self.assertEqual(list(m.r.keys()), [0, 2])
        self.assertEqual(list(m.r.values()), [m.x, m.y[1]])
        # But not add it back
        with self.assertRaisesRegex(
            KeyError, "Index '1' is not valid for indexed component 'r'"
        ):
            m.r[1] = m.x

    def test_reference_to_set(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 3, 5])
        m.r = Reference(m.I)
        self.assertEqual(len(m.r), 1)
        self.assertEqual(list(m.r.keys()), [None])
        self.assertEqual(list(m.r.values()), [m.I])
        self.assertIs(m.r[None], m.I)

        # Test that a referent Set containing None doesn't break the
        # None index
        m = ConcreteModel()
        m.I = Set(initialize=[1, 3, None, 5])
        m.r = Reference(m.I)
        self.assertEqual(len(m.r), 1)
        self.assertEqual(list(m.r.keys()), [None])
        self.assertEqual(list(m.r.values()), [m.I])
        self.assertIs(m.r[None], m.I)

    def test_is_reference(self):
        m = ConcreteModel()
        m.v0 = Var()
        m.v1 = Var([1, 2, 3])

        m.ref0 = Reference(m.v0)
        m.ref1 = Reference(m.v1)
        m.ref2 = Reference(m.v1[2])

        self.assertFalse(m.v0.is_reference())
        self.assertFalse(m.v1.is_reference())
        self.assertFalse(m.v1[2].is_reference())

        self.assertTrue(m.ref0.is_reference())
        self.assertTrue(m.ref1.is_reference())
        self.assertTrue(m.ref2.is_reference())

        unique_vars = list(v for v in m.component_objects(Var) if not v.is_reference())
        self.assertEqual(len(unique_vars), 2)

    def test_referent(self):
        m = ConcreteModel()
        m.v0 = Var()
        m.v2 = Var([1, 2, 3], ['a', 'b'])

        varlist = [m.v2[1, 'a'], m.v2[1, 'b']]

        vardict = {0: m.v0, 1: m.v2[1, 'a'], 2: m.v2[2, 'a'], 3: m.v2[3, 'a']}

        scalar_ref = Reference(m.v0)
        self.assertIs(scalar_ref.referent, m.v0)

        sliced_ref = Reference(m.v2[:, 'a'])
        referent = sliced_ref.referent
        self.assertIs(type(referent), IndexedComponent_slice)
        self.assertEqual(len(referent._call_stack), 1)
        call, info = referent._call_stack[0]
        self.assertEqual(call, IndexedComponent_slice.slice_info)
        self.assertIs(info[0], m.v2)
        self.assertEqual(info[1], {1: 'a'})  # Fixed
        self.assertEqual(info[2], {0: slice(None)})  # Sliced
        self.assertIs(info[3], None)  # Ellipsis

        list_ref = Reference(varlist)
        self.assertIs(list_ref.referent, varlist)

        dict_ref = Reference(vardict)
        self.assertIs(dict_ref.referent, vardict)

    def test_UnknownSetDimen(self):
        # Replicate the bug reported in #1928
        m = ConcreteModel()
        m.thinga = Set(initialize=['e1', 'e2', 'e3'])
        m.thingb = Set(initialize=[])
        m.v = Var(m.thinga | m.thingb)
        self.assertIs(m.v.dim(), UnknownSetDimen)
        with self.assertRaisesRegex(
            IndexError,
            'Slicing components relies on knowing the underlying set dimensionality',
        ):
            Reference(m.v)

    def test_contains_with_nonflattened(self):
        # test issue #1800
        _old_flatten = normalize_index.flatten
        try:
            normalize_index.flatten = False
            m = ConcreteModel()
            m.d1 = Set(initialize=[1, 2])
            m.d2 = Set(initialize=[('a', 1), ('b', 2)])
            m.v = Var(m.d2, m.d1)
            m.ref = Reference(m.v[:, 1])
            self.assertIn(('a', 1), m.ref)
            self.assertNotIn(('a', 10), m.ref)
        finally:
            normalize_index.flatten = _old_flatten

    def test_pprint_nonfinite_sets(self):
        m = ConcreteModel()
        m.v = Var(NonNegativeIntegers, dense=False)
        m.ref = Reference(m.v)
        buf = StringIO()
        m.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue().strip(),
            """
2 Var Declarations
    ref : Size=0, Index=NonNegativeIntegers, ReferenceTo=v
        Key : Lower : Value : Upper : Fixed : Stale : Domain
    v : Size=0, Index=NonNegativeIntegers
        Key : Lower : Value : Upper : Fixed : Stale : Domain

2 Declarations: v ref
""".strip(),
        )

        m.v[3]
        m.ref[5]
        buf = StringIO()
        m.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue().strip(),
            """
2 Var Declarations
    ref : Size=2, Index=NonNegativeIntegers, ReferenceTo=v
        Key : Lower : Value : Upper : Fixed : Stale : Domain
          3 :  None :  None :  None : False :  True :  Reals
          5 :  None :  None :  None : False :  True :  Reals
    v : Size=2, Index=NonNegativeIntegers
        Key : Lower : Value : Upper : Fixed : Stale : Domain
          3 :  None :  None :  None : False :  True :  Reals
          5 :  None :  None :  None : False :  True :  Reals

2 Declarations: v ref
""".strip(),
        )

    def test_pprint_nonfinite_sets_ctypeNone(self):
        # test issue #2039
        m = ConcreteModel()
        m.v = Var(NonNegativeIntegers, dense=False)
        m.ref = Reference(m.v, ctype=None)
        buf = StringIO()
        m.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue().strip(),
            """
1 Var Declarations
    v : Size=0, Index=NonNegativeIntegers
        Key : Lower : Value : Upper : Fixed : Stale : Domain

1 IndexedComponent Declarations
    ref : Size=0, Index=NonNegativeIntegers, ReferenceTo=v
        Key : Object

2 Declarations: v ref
""".strip(),
        )

        m.v[3]
        m.ref[5]
        buf = StringIO()
        m.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue().strip(),
            """
1 Var Declarations
    v : Size=2, Index=NonNegativeIntegers
        Key : Lower : Value : Upper : Fixed : Stale : Domain
          3 :  None :  None :  None : False :  True :  Reals
          5 :  None :  None :  None : False :  True :  Reals

1 IndexedComponent Declarations
    ref : Size=2, Index=NonNegativeIntegers, ReferenceTo=v
        Key : Object
          3 : <class 'pyomo.core.base.var.VarData'>
          5 : <class 'pyomo.core.base.var.VarData'>

2 Declarations: v ref
""".strip(),
        )

    def test_pprint_nested(self):
        m = ConcreteModel()

        @m.Block([1, 2])
        def b(b, i):
            b.x = Var([3, 4], bounds=(i, None))

        m.r = Reference(m.b[:].x[:])
        buf = StringIO()
        m.r.pprint(ostream=buf)
        self.assertEqual(
            buf.getvalue().strip(),
            """
r : Size=4, Index=ReferenceSet(b[:].x[:]), ReferenceTo=b[:].x[:]
    Key    : Lower : Value : Upper : Fixed : Stale : Domain
    (1, 3) :     1 :  None :  None : False :  True :  Reals
    (1, 4) :     1 :  None :  None : False :  True :  Reals
    (2, 3) :     2 :  None :  None : False :  True :  Reals
    (2, 4) :     2 :  None :  None : False :  True :  Reals
""".strip(),
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.appsi.cmodel import cmodel, cmodel_available
import pyomo.common.unittest as unittest
import math
from pyomo.contrib.fbbt.tests.test_interval import IntervalTestBase


@unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
class TestInterval(IntervalTestBase, unittest.TestCase):
    def setUp(self):
        super(TestInterval, self).setUp()
        self.add = cmodel.py_interval_add
        self.sub = cmodel.py_interval_sub
        self.mul = cmodel.py_interval_mul
        self.inv = cmodel.py_interval_inv
        self.div = cmodel.py_interval_div
        self.power = cmodel.py_interval_power
        self.exp = cmodel.py_interval_exp
        self.log = cmodel.py_interval_log
        self.log10 = cmodel.py_interval_log10
        self.sin = cmodel.py_interval_sin
        self.cos = cmodel.py_interval_cos
        self.tan = cmodel.py_interval_tan
        self.asin = cmodel.py_interval_asin
        self.acos = cmodel.py_interval_acos
        self.atan = cmodel.py_interval_atan
        self._inverse_power1 = cmodel._py_inverse_power1
        self._inverse_power2 = cmodel._py_inverse_power2


@unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
class TestCInterval(unittest.TestCase):
    def test_pow_with_inf(self):
        x_list = [
            0,
            -math.inf,
            math.inf,
            -3,
            3,
            -2,
            2,
            -1,
            1,
            -2.5,
            2.5,
            -0.5,
            0.5,
            -1.5,
            1.5,
        ]
        y_list = list(x_list)
        for x in x_list:
            for y in y_list:
                try:
                    expected = x**y
                    if type(expected) is complex:
                        expect_error = True
                        expected = None
                    else:
                        expect_error = False
                except ZeroDivisionError as e:
                    expected = None
                    expect_error = True
                if expect_error:
                    with self.assertRaises(ValueError):
                        cmodel._pow_with_inf(x, y)
                else:
                    got = cmodel._pow_with_inf(x, y)
                    self.assertAlmostEqual(expected, got)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
"""Testing for deprecated function."""

import logging
import sys

from importlib import import_module
from importlib.machinery import ModuleSpec
from io import StringIO

import pyomo.common
import pyomo.common.unittest as unittest

from pyomo.common import DeveloperError
from pyomo.common.deprecation import (
    deprecated,
    deprecation_warning,
    moved_module,
    relocated_module_attribute,
    MovedModuleFinder,
    MovedModuleLoader,
    RenamedClass,
    _import_object,
)
from pyomo.common.log import LoggingIntercept

logger = logging.getLogger('local')


class TestDeprecated(unittest.TestCase):
    """Tests for deprecated function decorator."""

    def test_deprecation_warning(self):
        DEP_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            deprecation_warning(None, version='1.2', remove_in='3.4')

        self.assertIn('DEPRECATED: This has been deprecated', DEP_OUT.getvalue())
        self.assertIn(
            '(deprecated in 1.2, will be removed in (or after) 3.4)',
            DEP_OUT.getvalue().replace('\n', ' '),
        )

        DEP_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            deprecation_warning("custom message here", version='1.2', remove_in='3.4')

        self.assertIn('DEPRECATED: custom message here', DEP_OUT.getvalue())
        self.assertIn(
            '(deprecated in 1.2, will be removed in (or after) 3.4)',
            DEP_OUT.getvalue().replace('\n', ' '),
        )

    def test_no_version_exception(self):
        with self.assertRaisesRegex(
            DeveloperError, r"@deprecated\(\): missing 'version' argument"
        ):

            @deprecated()
            def foo():
                pass

        with self.assertRaisesRegex(
            DeveloperError, r"@deprecated\(\): missing 'version' argument"
        ):

            @deprecated()
            class foo:
                pass

        # But no exception if the class can infer a version from the
        # __init__ (or __new__ or __new_member__)
        @deprecated()
        class foo:
            @deprecated(version="1.2")
            def __init__(self):
                pass

        self.assertIn('.. deprecated:: 1.2', foo.__doc__)

    def test_no_doc_string(self):
        # Note: No docstring, else nose replaces the function name with
        # the docstring in output.
        # """Test for deprecated function decorator."""
        @deprecated(version='test')
        def foo(bar='yeah'):
            logger.warning(bar)

        self.assertRegex(
            foo.__doc__,
            r'^DEPRECATED.\n\n.. deprecated:: test\n'
            r'   This function \(.*\.foo\) has been deprecated',
        )

        # Test the default argument
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo()
        # Test that the function produces output
        self.assertIn('yeah', FCN_OUT.getvalue())
        self.assertNotIn('DEPRECATED', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertRegex(
            DEP_OUT.getvalue().replace('\n', ' '),
            r'DEPRECATED: This function \(.*\.foo\) has been deprecated',
        )

        # Test that the function argument gets passed in
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo("custom")
        # Test that the function produces output
        self.assertNotIn('yeah', FCN_OUT.getvalue())
        self.assertIn('custom', FCN_OUT.getvalue())
        self.assertNotIn('DEPRECATED', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertRegex(
            DEP_OUT.getvalue().replace('\n', ' '),
            r'DEPRECATED: This function \(.*\.foo\) has been deprecated',
        )

    def test_with_doc_string(self):
        @deprecated(version='test')
        def foo(bar='yeah'):
            """Show that I am a good person.

            Because I document my public functions.

            """
            logger.warning(bar)

        self.assertRegex(
            foo.__doc__,
            r'I am a good person.\s+Because I document my public functions.\s+'
            r'.. deprecated:: test\n'
            r'   This function \(.*\.foo\) has been deprecated',
        )

        # Test the default argument
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo()
        # Test that the function produces output
        self.assertIn('yeah', FCN_OUT.getvalue())
        self.assertNotIn('DEPRECATED', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertRegex(
            DEP_OUT.getvalue().replace('\n', ' '),
            r'DEPRECATED: This function \(.*\.foo\) has been deprecated',
        )

        # Test that the function argument gets passed in
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo("custom")
        # Test that the function produces output
        self.assertNotIn('yeah', FCN_OUT.getvalue())
        self.assertIn('custom', FCN_OUT.getvalue())
        self.assertNotIn('DEPRECATED', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertRegex(
            DEP_OUT.getvalue().replace('\n', ' '),
            r'DEPRECATED: This function \(.*\.foo\) has been deprecated',
        )

    def test_with_custom_message(self):
        @deprecated('This is a custom message, too.', version='test')
        def foo(bar='yeah'):
            """Show that I am a good person.

            Because I document my public functions.

            """
            logger.warning(bar)

        self.assertIn('.. deprecated:: test\n   This is a custom message', foo.__doc__)
        self.assertIn('I am a good person.', foo.__doc__)

        # Test the default argument
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo()
        # Test that the function produces output
        self.assertIn('yeah', FCN_OUT.getvalue())
        self.assertNotIn('DEPRECATED', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertIn('DEPRECATED: This is a custom message', DEP_OUT.getvalue())

        # Test that the function argument gets passed in
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo("custom")
        # Test that the function produces output
        self.assertNotIn('yeah', FCN_OUT.getvalue())
        self.assertIn('custom', FCN_OUT.getvalue())
        self.assertNotIn('DEPRECATED', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertIn('DEPRECATED: This is a custom message', DEP_OUT.getvalue())

    def test_with_custom_logger(self):
        @deprecated('This is a custom message', logger='local', version='test')
        def foo(bar='yeah'):
            """Show that I am a good person.

            Because I document my public functions.

            """
            logger.warning(bar)

        self.assertIn('.. deprecated:: test\n   This is a custom message', foo.__doc__)
        self.assertIn('I am a good person.', foo.__doc__)

        # Test the default argument
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo()
        # Test that the function produces output
        self.assertIn('yeah', FCN_OUT.getvalue())
        self.assertIn('DEPRECATED: This is a custom message', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertNotIn('DEPRECATED:', DEP_OUT.getvalue())

        # Test that the function argument gets passed in
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo("custom")
        # Test that the function produces output
        self.assertNotIn('yeah', FCN_OUT.getvalue())
        self.assertIn('custom', FCN_OUT.getvalue())
        self.assertIn('DEPRECATED: This is a custom message', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertNotIn('DEPRECATED:', DEP_OUT.getvalue())

    def test_with_class(self):
        @deprecated(version='test')
        class foo:
            def __init__(self):
                logger.warning('yeah')

        self.assertIs(type(foo), type)
        self.assertRegex(
            foo.__doc__,
            r'.. deprecated:: test\n   This class \(.*\.foo\) has been deprecated',
        )

        # Test the default argument
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo()
        # Test that the function produces output
        self.assertIn('yeah', FCN_OUT.getvalue())
        self.assertNotIn('DEPRECATED', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertRegex(
            DEP_OUT.getvalue().replace('\n', ' '),
            r'DEPRECATED: This class \(.*\.foo\) has been deprecated.*'
            r'\(deprecated in test\)',
        )

    def test_with_method(self):
        class foo:
            def __init__(self):
                pass

            @deprecated(version='test')
            def bar(self):
                logger.warning('yeah')

        self.assertRegex(
            foo.bar.__doc__,
            r'.. deprecated:: test\n'
            r'   This function \(.*\.foo\.bar\) has been deprecated',
        )

        # Test the default argument
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo().bar()
        # Test that the function produces output
        self.assertIn('yeah', FCN_OUT.getvalue())
        self.assertNotIn('DEPRECATED', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertRegex(
            DEP_OUT.getvalue().replace('\n', ' '),
            r'DEPRECATED: This function \(.*\.foo\.bar\) has been deprecated.*'
            r'\(deprecated in test\)',
        )

    def test_with_remove_in(self):
        class foo:
            def __init__(self):
                pass

            @deprecated(version='1.2', remove_in='3.4')
            def bar(self):
                logger.warning('yeah')

        self.assertRegex(
            foo.bar.__doc__,
            r'.. deprecated:: 1.2\n'
            r'   This function \(.*\.foo\.bar\) has been deprecated.*'
            r'\(will be removed in \(or after\) 3.4\)',
        )

        # Test the default argument
        DEP_OUT = StringIO()
        FCN_OUT = StringIO()
        with LoggingIntercept(DEP_OUT, 'pyomo'):
            with LoggingIntercept(FCN_OUT, 'local'):
                foo().bar()
        # Test that the function produces output
        self.assertIn('yeah', FCN_OUT.getvalue())
        self.assertNotIn('DEPRECATED', FCN_OUT.getvalue())
        # Test that the deprecation warning was logged
        self.assertRegex(
            DEP_OUT.getvalue().replace('\n', ' '),
            r'DEPRECATED: This function \(.*\.foo\.bar\) has been deprecated.*'
            r'\(deprecated in 1.2, will be removed in \(or after\) 3.4\)',
        )


class Bar:
    data = 21


relocated_module_attribute('myFoo', 'pyomo.common.tests.relocated.Bar', 'test')


class TestRelocated(unittest.TestCase):
    def test_relocated_class(self):
        # Before we test multiple relocated objects, verify that it will
        # handle the import of a new module
        warning = (
            "DEPRECATED: the 'myFoo' class has been moved to "
            "'pyomo.common.tests.relocated.Bar'"
        )
        OUT = StringIO()
        with LoggingIntercept(OUT, 'pyomo'):
            from pyomo.common.tests.test_deprecated import myFoo
        self.assertEqual(myFoo.data, 42)
        self.assertIn(warning, OUT.getvalue().replace('\n', ' '))

        from pyomo.common.tests import relocated

        self.assertNotIn('Foo', dir(relocated))
        self.assertNotIn('Foo_2', dir(relocated))

        warning = (
            "DEPRECATED: the 'Foo_2' class has been moved to "
            "'pyomo.common.tests.relocated.Bar'"
        )

        OUT = StringIO()
        with LoggingIntercept(OUT, 'pyomo'):
            self.assertIs(relocated.Foo_2, relocated.Bar)
            self.assertEqual(relocated.Foo_2.data, 42)
        self.assertIn(warning, OUT.getvalue().replace('\n', ' '))

        self.assertNotIn('Foo', dir(relocated))
        self.assertIn('Foo_2', dir(relocated))
        self.assertIs(relocated.Foo_2, relocated.Bar)

        warning = (
            "DEPRECATED: the 'Foo' class has been moved to "
            "'pyomo.common.tests.test_deprecated.Bar'"
        )

        OUT = StringIO()
        with LoggingIntercept(OUT, 'pyomo'):
            from pyomo.common.tests.relocated import Foo

            self.assertEqual(Foo.data, 21)
        self.assertIn(warning, OUT.getvalue().replace('\n', ' '))

        self.assertIn('Foo', dir(relocated))
        self.assertIn('Foo_2', dir(relocated))
        self.assertIs(relocated.Foo, Bar)

        # Note that relocated defines a __getattr__, which changes how
        # attribute processing is handled in python 3.7+
        with self.assertRaisesRegex(
            AttributeError,
            "(?:module 'pyomo.common.tests.relocated')|"
            "(?:'module' object) has no attribute 'Baz'",
        ):
            relocated.Baz.data
        self.assertEqual(relocated.Foo_3, '_3')

        with self.assertRaisesRegex(
            AttributeError,
            "(?:module 'pyomo.common.tests.test_deprecated')|"
            "(?:'module' object) has no attribute 'Baz'",
        ):
            sys.modules[__name__].Baz.data

    def test_relocated_message(self):
        with LoggingIntercept() as LOG:
            self.assertIs(
                _import_object(
                    'oldName',
                    'pyomo.common.tests.test_deprecated.logger',
                    'TBD',
                    None,
                    None,
                ),
                logger,
            )
        self.assertRegex(
            LOG.getvalue().replace('\n', ' '),
            "DEPRECATED: the 'oldName' attribute has been moved to "
            "'pyomo.common.tests.test_deprecated.logger'",
        )

        with LoggingIntercept() as LOG:
            self.assertIs(
                _import_object(
                    'oldName',
                    'pyomo.common.tests.test_deprecated._import_object',
                    'TBD',
                    None,
                    None,
                ),
                _import_object,
            )
        self.assertRegex(
            LOG.getvalue().replace('\n', ' '),
            "DEPRECATED: the 'oldName' function has been moved to "
            "'pyomo.common.tests.test_deprecated._import_object'",
        )

        with LoggingIntercept() as LOG:
            self.assertIs(
                _import_object(
                    'oldName',
                    'pyomo.common.tests.test_deprecated.TestRelocated',
                    'TBD',
                    None,
                    None,
                ),
                TestRelocated,
            )
        self.assertRegex(
            LOG.getvalue().replace('\n', ' '),
            "DEPRECATED: the 'oldName' class has been moved to "
            "'pyomo.common.tests.test_deprecated.TestRelocated'",
        )

    def test_relocated_module(self):
        with LoggingIntercept() as LOG:
            # Can import attributes defined only in the new module
            from pyomo.common.tests.relo_mod import ReloClass
        self.assertRegex(
            LOG.getvalue().replace('\n', ' '),
            r"DEPRECATED: The 'pyomo\.common\.tests\.relo_mod' module has "
            r"been moved to 'pyomo\.common\.tests\.relo_mod_new'. Please "
            r"update your import. \(deprecated in 1\.2\) \(called from "
            r".*test_deprecated\.py",
        )
        with LoggingIntercept() as LOG:
            # Second import: no warning
            import pyomo.common.tests.relo_mod as relo
        self.assertEqual(LOG.getvalue(), '')
        import pyomo.common.tests.relo_mod_new as relo_new

        self.assertIs(relo, relo_new)
        self.assertEqual(relo.RELO_ATTR, 42)
        self.assertIs(ReloClass, relo_new.ReloClass)


class TestRenamedClass(unittest.TestCase):
    def test_renamed(self):
        class NewClass:
            attr = 'NewClass'

        class NewClassSubclass(NewClass):
            pass

        # The deprecated class does not generate a warning
        out = StringIO()
        with LoggingIntercept(out):

            class DeprecatedClass(metaclass=RenamedClass):
                __renamed__new_class__ = NewClass
                __renamed__version__ = 'X.y'

        self.assertEqual(out.getvalue(), "")

        # Inheriting from the deprecated class generates the warning
        out = StringIO()
        with LoggingIntercept(out):

            class DeprecatedClassSubclass(DeprecatedClass):
                attr = 'DeprecatedClassSubclass'

        self.assertRegex(
            out.getvalue().replace("\n", " ").strip(),
            r"^DEPRECATED: Declaring class 'DeprecatedClassSubclass' "
            r"derived from 'DeprecatedClass'.  "
            r"The class 'DeprecatedClass' has been renamed to 'NewClass'.  "
            r"\(deprecated in X.y\) \(called from [^\)]*\)$",
        )

        # Inheriting from a class derived from the deprecated class does
        # not generate a warning
        out = StringIO()
        with LoggingIntercept(out):

            class otherClass:
                pass

            class DeprecatedClassSubSubclass(DeprecatedClassSubclass, otherClass):
                attr = 'DeprecatedClassSubSubclass'

        self.assertEqual(out.getvalue(), "")

        #
        # Test class creation
        #

        out = StringIO()
        with LoggingIntercept(out):
            newclass = NewClass()
            newclasssubclass = NewClassSubclass()
        self.assertEqual(out.getvalue(), "")

        out = StringIO()
        with LoggingIntercept(out):
            deprecatedclass = DeprecatedClass()
        self.assertRegex(
            out.getvalue().replace("\n", " ").strip(),
            r"^DEPRECATED: Instantiating class 'DeprecatedClass'.  "
            r"The class 'DeprecatedClass' has been renamed to 'NewClass'.  "
            r"\(deprecated in X.y\) \(called from [^\)]*\)$",
        )

        # Instantiating a class derived from the deprecaed class does
        # not generate a warning (the warning is generated when the
        # class is declared)
        out = StringIO()
        with LoggingIntercept(out):
            deprecatedsubclass = DeprecatedClassSubclass()
            deprecatedsubsubclass = DeprecatedClassSubSubclass()
        self.assertEqual(out.getvalue(), "")

        #
        # Test isinstance
        #
        out = StringIO()
        with LoggingIntercept(out):
            self.assertIsInstance(deprecatedsubclass, NewClass)
            self.assertIsInstance(deprecatedsubsubclass, NewClass)
        self.assertEqual(out.getvalue(), "")

        for obj in (
            newclass,
            newclasssubclass,
            deprecatedclass,
            deprecatedsubclass,
            deprecatedsubsubclass,
        ):
            out = StringIO()
            with LoggingIntercept(out):
                self.assertIsInstance(obj, DeprecatedClass)
            self.assertRegex(
                out.getvalue().replace("\n", " ").strip(),
                r"^DEPRECATED: Checking type relative to 'DeprecatedClass'.  "
                r"The class 'DeprecatedClass' has been renamed to 'NewClass'."
                r"  \(deprecated in X.y\) \(called from [^\)]*\)$",
            )

        #
        # Test issubclass
        #

        out = StringIO()
        with LoggingIntercept(out):
            self.assertTrue(issubclass(DeprecatedClass, NewClass))
            self.assertTrue(issubclass(DeprecatedClassSubclass, NewClass))
            self.assertTrue(issubclass(DeprecatedClassSubSubclass, NewClass))
        self.assertEqual(out.getvalue(), "")

        for cls in (
            NewClass,
            NewClassSubclass,
            DeprecatedClass,
            DeprecatedClassSubclass,
            DeprecatedClassSubSubclass,
        ):
            out = StringIO()
            with LoggingIntercept(out):
                self.assertTrue(issubclass(cls, DeprecatedClass))
            self.assertRegex(
                out.getvalue().replace("\n", " ").strip(),
                r"^DEPRECATED: Checking type relative to 'DeprecatedClass'.  "
                r"The class 'DeprecatedClass' has been renamed to 'NewClass'."
                r"  \(deprecated in X.y\) \(called from [^\)]*\)$",
            )

        #
        # Test class attributes
        #
        self.assertEqual(newclass.attr, 'NewClass')
        self.assertEqual(newclasssubclass.attr, 'NewClass')
        self.assertEqual(deprecatedclass.attr, 'NewClass')
        self.assertEqual(deprecatedsubclass.attr, 'DeprecatedClassSubclass')
        self.assertEqual(deprecatedsubsubclass.attr, 'DeprecatedClassSubSubclass')
        self.assertEqual(NewClass.attr, 'NewClass')
        self.assertEqual(NewClassSubclass.attr, 'NewClass')
        self.assertEqual(DeprecatedClass.attr, 'NewClass')
        self.assertEqual(DeprecatedClassSubclass.attr, 'DeprecatedClassSubclass')
        self.assertEqual(DeprecatedClassSubSubclass.attr, 'DeprecatedClassSubSubclass')

    def test_renamed_errors(self):
        class NewClass:
            pass

        with self.assertRaisesRegex(
            TypeError,
            "Declaring class 'DeprecatedClass' using the "
            "RenamedClass metaclass, but without specifying the "
            "__renamed__new_class__ class attribute",
        ):

            class DeprecatedClass(metaclass=RenamedClass):
                __renamed_new_class__ = NewClass

        with self.assertRaisesRegex(
            DeveloperError,
            "Declaring class 'DeprecatedClass' using the "
            "RenamedClass metaclass, but without specifying the "
            "__renamed__version__ class attribute",
            normalize_whitespace=True,
        ):

            class DeprecatedClass(metaclass=RenamedClass):
                __renamed__new_class__ = NewClass


class TestMoved(unittest.TestCase):
    def test_finder(self):
        mod_name = 'pyomo.common.deprecation_tester'
        finder = MovedModuleFinder()
        self.assertNotIn(mod_name, finder.mapping)
        self.assertIsNone(finder.find_spec(mod_name, pyomo.common.__path__))

        moved_module(mod_name, __name__, version='1.2.3')
        try:
            self.assertIn(mod_name, finder.mapping)
            spec = finder.find_spec(mod_name, pyomo.common.__path__)
            self.assertIs(type(spec), ModuleSpec)
            self.assertEqual(spec.name, mod_name)
            self.assertIs(type(spec.loader), MovedModuleLoader)
            self.assertEqual(spec.origin, __file__)
        finally:
            del finder.mapping[mod_name]

    def test_declaration(self):
        try:
            _old = 'pyomo.common.tests.old_moved'
            _new = 'pyomo.common.tests.moved'
            # 1st registration is OK
            N = len(MovedModuleFinder.mapping)
            self.assertNotIn(_old, MovedModuleFinder.mapping)
            moved_module(_old, _new, version='1.2')
            self.assertIn(_old, MovedModuleFinder.mapping)
            self.assertEqual(N + 1, len(MovedModuleFinder.mapping))
            # duplicate registration is OK
            moved_module(_old, _new, version='1.2')
            self.assertIn(_old, MovedModuleFinder.mapping)
            self.assertEqual(N + 1, len(MovedModuleFinder.mapping))
            _conflict = 'pyomo.something.else'
            with self.assertRaisesRegex(
                RuntimeError,
                "(?s)Duplicate module alias declaration.\n"
                f"\toriginal: {_old} -> {_new}\n"
                f"\tconflict: {_old} -> {_conflict}\n",
            ):
                moved_module(_old, _conflict, version='1.2')
            self.assertIn(_old, MovedModuleFinder.mapping)
            self.assertEqual(N + 1, len(MovedModuleFinder.mapping))
        finally:
            del MovedModuleFinder.mapping[_old]

    def test_loader(self):
        mod_name = 'pyomo.common.deprecation_tester'
        try:
            moved_module(mod_name, __name__, version='1.2.3')
            with LoggingIntercept() as LOG:
                import pyomo.common.deprecation_tester
            self.assertRegex(
                LOG.getvalue().replace('\n', ' ').strip(),
                "DEPRECATED: The 'pyomo.common.deprecation_tester' module has been "
                "moved to 'pyomo.common.tests.test_deprecated'. Please update your "
                r"import.  \(deprecated in 1.2.3\) \(called from [^)]+\)",
            )
            self.assertIs(pyomo.common.deprecation_tester.TestMoved, TestMoved)
        finally:
            del MovedModuleFinder.mapping[mod_name]

        try:
            moved_module(mod_name, __name__, msg=None, version='1.2.3')
            with LoggingIntercept() as LOG:
                import pyomo.common.deprecation_tester
            self.assertEqual(LOG.getvalue(), "")
            self.assertIs(pyomo.common.deprecation_tester.TestMoved, TestMoved)
        finally:
            del MovedModuleFinder.mapping[mod_name]

        try:
            moved_module(
                'pyomo.common.tests.old_moved',
                'pyomo.common.tests.moved',
                version='1.2',
            )
            self.assertNotIn('pyomo.common.tests.moved', sys.modules)
            self.assertNotIn('pyomo.common.tests.old_moved', sys.modules)
            with LoggingIntercept() as LOG:
                import pyomo.common.tests.old_moved
            self.assertRegex(
                LOG.getvalue().replace('\n', ' ').strip(),
                "DEPRECATED: The 'pyomo.common.tests.old_moved' module has been "
                "moved to 'pyomo.common.tests.moved'. Please update your "
                r"import.  \(deprecated in 1.2\) \(called from [^)]+\)",
            )
            self.assertIn('pyomo.common.tests.moved', sys.modules)
            self.assertIn('pyomo.common.tests.old_moved', sys.modules)
            self.assertIs(
                sys.modules['pyomo.common.tests.moved'],
                sys.modules['pyomo.common.tests.old_moved'],
            )
        finally:
            del MovedModuleFinder.mapping['pyomo.common.tests.old_moved']
            del sys.modules['pyomo.common.tests.old_moved']
            del sys.modules['pyomo.common.tests.moved']

    def test_archive_importable(self):
        import pyomo.environ

        # Check that all modules in the _archive directory are importable.
        for old_name, info in MovedModuleFinder.mapping.items():
            if '._archive.' in info.new_name:
                with LoggingIntercept() as LOG:
                    m = import_module(info.old_name)
                self.assertIn('DEPRECATED', LOG.getvalue())
                # We expect every module in _archive to be deprecated
                # (and to state that in the module docstring):
                self.assertIn('deprecated', m.__doc__)
                self.assertEqual(m.__name__, info.new_name)
                # Remove these modules from sys.modules (some other
                # modules have tests for deprecation paths that rely on
                # these modules not having already been imported)
                del sys.modules[info.old_name]
                del sys.modules[info.new_name]


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

logger = logging.getLogger(__name__)

from pyomo.common.dependencies import attempt_import
from pyomo.common.errors import ApplicationError

from pyomo.contrib import appsi
import pyomo.contrib.alternative_solutions.aos_utils as aos_utils
from pyomo.contrib.alternative_solutions import Solution


def gurobi_generate_solutions(
    model,
    *,
    num_solutions=10,
    rel_opt_gap=None,
    abs_opt_gap=None,
    solver_options={},
    tee=False,
):
    """
    Finds alternative optimal solutions for discrete variables using Gurobi's
    built-in Solution Pool capability.

    This method defaults to the optimality-enforced discovery method with PoolSearchMode = 2.
    There are two other options, standard single optimal solution (PoolSearchMode = 0) and
    best-effort discovery with no guarantees (PoolSearchMode = 1). Please consult the Gurobi
    documentation on PoolSearchMode for details on impact on Gurobi results.
    Changes to this mode can be made by included PoolSearchMode set to the intended value
    in solver_options.

    Parameters
    ----------
    model : ConcreteModel
        A concrete Pyomo model.
    num_solutions : int
        The maximum number of solutions to generate. This parameter maps to
        the PoolSolutions parameter in Gurobi.
    rel_opt_gap : non-negative float or None
        The relative optimality gap for allowable alternative solutions.
        None implies that there is no limit on the relative optimality gap
        (i.e. that any feasible solution can be considered by Gurobi).
        This parameter maps to the PoolGap parameter in Gurobi.
    abs_opt_gap : non-negative float or None
        The absolute optimality gap for allowable alternative solutions.
        None implies that there is no limit on the absolute optimality gap
        (i.e. that any feasible solution can be considered by Gurobi).
        This parameter maps to the PoolGapAbs parameter in Gurobi.
    solver_options : dict
        Solver option-value pairs to be passed to the Gurobi solver.
    tee : boolean
        Boolean indicating that the solver output should be displayed.

    Returns
    -------
    solutions
        A list of Solution objects.  [Solution]
    """
    #
    # Setup gurobi
    #
    opt = appsi.solvers.Gurobi()
    if not opt.available():
        raise ApplicationError("Solver (gurobi) not available")

    assert num_solutions >= 1, "num_solutions must be positive integer"
    if num_solutions == 1:
        logger.warning("Running alternative_solutions method to find only 1 solution!")

    opt.config.stream_solver = tee
    opt.config.load_solution = False
    opt.gurobi_options["PoolSolutions"] = num_solutions
    if rel_opt_gap is not None:
        opt.gurobi_options["PoolGap"] = rel_opt_gap
    if abs_opt_gap is not None:
        opt.gurobi_options["PoolGapAbs"] = abs_opt_gap
    for parameter, value in solver_options.items():
        opt.gurobi_options[parameter] = value
    if "PoolSearchMode" not in opt.gurobi_options:
        opt.gurobi_options["PoolSearchMode"] = 2
    elif opt.gurobi_options["PoolSearchMode"] == 0:
        logger.warning(
            "Running gurobi_solnpool with PoolSearchMode=0, this is single search mode and not the intended use case for gurobi_generate_solutions"
        )
    elif opt.gurobi_options["PoolSearchMode"] == 1:
        logger.warning(
            "Running gurobi_solnpool with PoolSearchMode=1, best effort search may lead to unexpected behavior"
        )
    #
    # Run gurobi
    #
    results = opt.solve(model)
    condition = results.termination_condition
    if not (condition == appsi.base.TerminationCondition.optimal):
        raise ApplicationError(
            "Model cannot be solved, " "TerminationCondition = {}"
        ).format(condition.value)
    #
    # Collect solutions
    #
    solution_count = opt.get_model_attr("SolCount")
    variables = aos_utils.get_model_variables(model, include_fixed=True)
    solutions = []
    for i in range(solution_count):
        #
        # Load the i-th solution into the model
        #
        results.solution_loader.load_vars(solution_number=i)
        #
        # Pull the solution from the model into a Solution object,
        # and append to our list of solutions
        #
        solutions.append(Solution(model, variables))

    return solutions



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from .triangularize import block_triangularize
from .matching import maximum_matching
from .interface import IncidenceGraphInterface, get_bipartite_incidence_graph
from .scc_solver import (
    generate_strongly_connected_components,
    solve_strongly_connected_components,
)
from .incidence import get_incident_variables
from .config import IncidenceMethod

#
# declare deprecation paths for removed modules
#
from pyomo.common.deprecation import moved_module

moved_module(
    "pyomo.contrib.incidence_analysis.util",
    "pyomo.contrib.incidence_analysis.scc_solver",
    version='6.5.0',
    msg=(
        "The 'pyomo.contrib.incidence_analysis.util' module has been moved to "
        "'pyomo.contrib.incidence_analysis.scc_solver'. However, we recommend "
        "importing this functionality (e.g. solve_strongly_connected_components) "
        "directly from 'pyomo.contrib.incidence_analysis'."
    ),
)
del moved_module



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
from io import StringIO
from pyomo.common.log import LoggingIntercept
from pyomo.common.tee import capture_output
from pyomo.repn.tests.lp_diff import lp_diff

_baseline = r"""\* Source Pyomo model name=unknown *\

min
x2:
+ [
+2 x1 ^ 2
] / 2

s.t.

c_l_x3_:
+1 x1
>= 1

c_e_ONE_VAR_CONSTANT:
ONE_VAR_CONSTANT = 1.0

bounds
    -inf <= x1 <= +inf
end
"""


def _check_log_and_out(LOG, OUT, offset, msg=None):
    sys.stdout.flush()
    sys.stderr.flush()
    msg = str(msg) + ': ' if msg else ''
    if LOG.getvalue():
        raise RuntimeError(
            "FAIL: %sMessage logged to the Logger:\n>>>\n%s<<<" % (msg, LOG.getvalue())
        )

    if OUT.getvalue():
        raise RuntimeError(
            "FAIL: %sMessage sent to stdout/stderr:\n>>>\n%s<<<" % (msg, OUT.getvalue())
        )


def import_pyomo_environ():
    with LoggingIntercept() as LOG, capture_output(capture_fd=True) as OUT:
        import pyomo.environ as pyo

        globals()['pyo'] = pyo
    _check_log_and_out(LOG, OUT, 0)


def run_writer_test():
    with LoggingIntercept() as LOG, capture_output(capture_fd=True) as OUT:
        # Enumerate the writers...
        from pyomo.opt import WriterFactory

        info = []
        for writer in sorted(WriterFactory):
            info.append("  %s: %s" % (writer, WriterFactory.doc(writer)))
            _check_log_and_out(LOG, OUT, 10, writer)

    print("Pyomo Problem Writers")
    print("---------------------")
    print('\n'.join(info))

    with LoggingIntercept() as LOG, capture_output(capture_fd=True) as OUT:
        # Test a writer
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.c = pyo.Constraint(expr=m.x >= 1)
        m.o = pyo.Objective(expr=m.x**2)

        from pyomo.common.tempfiles import TempfileManager

        with TempfileManager:
            fname = TempfileManager.create_tempfile(suffix='pyomo.lp_v1')
            m.write(fname, format='lp_v1')
            with open(fname, 'r') as FILE:
                data = FILE.read()

    base, test = lp_diff(_baseline, data)
    if base != test:
        print(
            "Result did not match baseline.\nRESULT:\n%s\nBASELINE:\n%s"
            % (data, _baseline)
        )
        print(data.strip().splitlines())
        print(_baseline.strip().splitlines())
        sys.exit(2)

    _check_log_and_out(LOG, OUT, 10)


def run_solverfactory_test():
    skip_solvers = {'py', 'xpress', '_xpress_shell', '_mock_xpress'}

    with LoggingIntercept() as LOG, capture_output(capture_fd=True) as OUT:
        info = []
        for solver in sorted(pyo.SolverFactory):
            _doc = pyo.SolverFactory.doc(solver)
            if _doc is not None and 'DEPRECATED' in _doc:
                _avail = 'DEPR'
            elif solver in skip_solvers:
                _avail = 'SKIP'
            else:
                _avail = str(pyo.SolverFactory(solver).available(False))
            info.append("   %s(%s): %s" % (solver, _avail, _doc))
            # _check_log_and_out(LOG, OUT, 20, solver)

        glpk = pyo.SolverFactory('glpk')

    print("")
    print("Pyomo Solvers")
    print("-------------")
    print("\n".join(info))

    if type(glpk.available(False)) != bool:
        print("Solver glpk.available() did not return bool")
        sys.exit(3)

    _check_log_and_out(LOG, OUT, 20)


def run_transformationfactory_test():
    with LoggingIntercept() as LOG, capture_output(capture_fd=True) as OUT:
        info = []
        for t in sorted(pyo.TransformationFactory):
            _doc = pyo.TransformationFactory.doc(t)
            info.append("   %s: %s" % (t, _doc))
            if 'DEPRECATED' not in _doc:
                pyo.TransformationFactory(t)
            _check_log_and_out(LOG, OUT, 30, t)

        bigm = pyo.TransformationFactory('gdp.bigm')

    print("")
    print("Pyomo Transformations")
    print("---------------------")
    print('\n'.join(info))

    if not isinstance(bigm, pyo.Transformation):
        print("TransformationFactory(gdp.bigm) did not return a transformation")
        sys.exit(4)

    _check_log_and_out(LOG, OUT, 30)


if __name__ == '__main__':
    # Run some basic tests: map all errors / warnings / exceptions to a
    # non-zero process return code, so that these tests can be run
    # outside of any test harness
    try:
        import_pyomo_environ()
        run_writer_test()
        run_solverfactory_test()
        run_transformationfactory_test()
    except:
        et, e, tb = sys.exc_info()
        print(str(e))
        sys.exit(1)
    sys.exit(0)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
from pyomo.contrib.piecewise.transform.piecewise_linear_transformation_base import (
    PiecewiseLinearTransformationBase,
)
from pyomo.core import Constraint, NonNegativeIntegers, Suffix, Var
from pyomo.core.base import TransformationFactory
from pyomo.gdp import Disjunction
from pyomo.common.errors import DeveloperError


@TransformationFactory.register(
    "contrib.piecewise.nested_inner_repn_gdp",
    doc="""
    Represent a piecewise linear function by using a nested GDP to determine
    which polytope a point is in, then representing it as a convex combination
    of extreme points, with multipliers "local" to that particular polytope,
    i.e., not shared with neighbors. This formulation has linearly many Boolean
    variables, though up to variable substitution, it has logarithmically many.
    """,
)
class NestedInnerRepresentationGDPTransformation(PiecewiseLinearTransformationBase):
    """
    Represent a piecewise linear function by using a nested GDP to determine
    which polytope a point is in, then representing it as a convex combination
    of extreme points, with multipliers "local" to that particular polytope,
    i.e., not shared with neighbors. This method of formulating the piecewise
    linear function imposes no restrictions on the family of polytopes. Note
    that this is NOT a logarithmic formulation - it has linearly many Boolean
    variables.  However, it is inspired by the disaggregated logarithmic
    formulation of [1]. Up to variable substitution, the amount of Boolean
    variables is logarithmic, as in [1].

    References
    ----------
    [1] J.P. Vielma, S. Ahmed, and G. Nemhauser, "Mixed-integer models
        for nonseparable piecewise-linear optimization: unifying framework
        and extensions," Operations Research, vol. 58, no. 2, pp. 305-315,
        2010.
    """

    CONFIG = PiecewiseLinearTransformationBase.CONFIG()
    _transformation_name = "pw_linear_nested_inner_repn"

    # Implement to use PiecewiseLinearTransformationBase. This function returns the Var
    # that replaces the transformed piecewise linear expr
    def _transform_pw_linear_expr(self, pw_expr, pw_linear_func, transformation_block):
        # Get a new Block() in transformation_block.transformed_functions, which
        # is a Block(Any)
        transBlock = transformation_block.transformed_functions[
            len(transformation_block.transformed_functions)
        ]

        substitute_var = transBlock.substitute_var = Var()
        pw_linear_func.map_transformation_var(pw_expr, substitute_var)
        transBlock.substitute_var_lb = float("inf")
        transBlock.substitute_var_ub = -float("inf")

        choices = list(zip(pw_linear_func._simplices, pw_linear_func._linear_functions))

        # If there was only one choice, don't bother making a disjunction, just
        # use the linear function directly (but still use the substitute_var for
        # consistency).
        if len(choices) == 1:
            _, linear_func = choices[0]  # simplex isn't important in this case
            linear_func_expr = linear_func(*pw_expr.args)
            transBlock.set_substitute = Constraint(
                expr=substitute_var == linear_func_expr
            )
            transBlock.substitute_var_lb, transBlock.substitute_var_ub = (
                compute_bounds_on_expr(linear_func_expr)
            )
        else:
            # Add the disjunction
            transBlock.disj = self._get_disjunction(
                choices, transBlock, pw_expr, pw_linear_func, transBlock
            )

        # Set bounds as determined when setting up the disjunction
        if transBlock.substitute_var_lb < float("inf"):
            transBlock.substitute_var.setlb(transBlock.substitute_var_lb)
        if transBlock.substitute_var_ub > -float("inf"):
            transBlock.substitute_var.setub(transBlock.substitute_var_ub)

        return substitute_var

    # Recursively form the Disjunctions and Disjuncts. This shouldn't blow up
    # the stack, since the whole point is that we'll only go logarithmically
    # many calls deep.
    def _get_disjunction(
        self, choices, parent_block, pw_expr, pw_linear_func, root_block
    ):
        size = len(choices)

        # Our base cases will be 3 and 2, since it would be silly to construct
        # a Disjunction containing only one Disjunct. We can ensure that size
        # is never 1 unless it was only passed a single choice from the start,
        # which we can handle before calling.
        if size > 3:
            half = size // 2  # (integer divide)
            # This tree will be slightly heavier on the right side
            choices_l = choices[:half]
            choices_r = choices[half:]

            @parent_block.Disjunct()
            def d_l(b):
                b.inner_disjunction_l = self._get_disjunction(
                    choices_l, b, pw_expr, pw_linear_func, root_block
                )

            @parent_block.Disjunct()
            def d_r(b):
                b.inner_disjunction_r = self._get_disjunction(
                    choices_r, b, pw_expr, pw_linear_func, root_block
                )

            return Disjunction(expr=[parent_block.d_l, parent_block.d_r])
        elif size == 3:
            # Let's stay heavier on the right side for consistency. So the left
            # Disjunct will be the one to contain constraints, rather than a
            # Disjunction
            @parent_block.Disjunct()
            def d_l(b):
                simplex, linear_func = choices[0]
                self._set_disjunct_block_constraints(
                    b, simplex, linear_func, pw_expr, pw_linear_func, root_block
                )

            @parent_block.Disjunct()
            def d_r(b):
                b.inner_disjunction_r = self._get_disjunction(
                    choices[1:], b, pw_expr, pw_linear_func, root_block
                )

            return Disjunction(expr=[parent_block.d_l, parent_block.d_r])
        elif size == 2:
            # In this case both sides are regular Disjuncts
            @parent_block.Disjunct()
            def d_l(b):
                simplex, linear_func = choices[0]
                self._set_disjunct_block_constraints(
                    b, simplex, linear_func, pw_expr, pw_linear_func, root_block
                )

            @parent_block.Disjunct()
            def d_r(b):
                simplex, linear_func = choices[1]
                self._set_disjunct_block_constraints(
                    b, simplex, linear_func, pw_expr, pw_linear_func, root_block
                )

            return Disjunction(expr=[parent_block.d_l, parent_block.d_r])
        else:
            raise DeveloperError(
                "Unreachable: 1 or 0 choices were passed to "
                "_get_disjunction in nested_inner_repn.py."
            )

    def _set_disjunct_block_constraints(
        self, b, simplex, linear_func, pw_expr, pw_linear_func, root_block
    ):
        # Define the lambdas sparsely like in the normal inner repn,
        # only the first few will participate in constraints
        b.lambdas = Var(NonNegativeIntegers, dense=False, bounds=(0, 1))

        # Get the extreme points to add up
        extreme_pts = []
        for idx in simplex:
            extreme_pts.append(pw_linear_func._points[idx])

        # Constrain sum(lambda_i) = 1
        b.convex_combo = Constraint(
            expr=sum(b.lambdas[i] for i in range(len(extreme_pts))) == 1
        )
        linear_func_expr = linear_func(*pw_expr.args)

        # Make the substitute Var equal the PWLE
        b.set_substitute = Constraint(
            expr=root_block.substitute_var == linear_func_expr
        )

        # Widen the variable bounds to those of this linear func expression
        lb, ub = compute_bounds_on_expr(linear_func_expr)
        if lb is not None and lb < root_block.substitute_var_lb:
            root_block.substitute_var_lb = lb
        if ub is not None and ub > root_block.substitute_var_ub:
            root_block.substitute_var_ub = ub

        # Constrain x = \sum \lambda_i v_i
        @b.Constraint(range(pw_expr.nargs()))  # dimension
        def linear_combo(d, i):
            return pw_expr.args[i] == sum(
                d.lambdas[j] * pt[i] for j, pt in enumerate(extreme_pts)
            )

        # Mark the lambdas as local in order to prevent disagreggating multiple
        # times in the hull transformation
        b.LocalVars = Suffix(direction=Suffix.LOCAL)
        b.LocalVars[b] = [v for v in b.lambdas.values()]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
import math
from pyomo.core.base.block import BlockData
from pyomo.common.collections import ComponentSet
from pyomo.core.base.var import Var
from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
from pyomo.core.expr.calculus.diff_with_pyomo import reverse_sd
import logging

logger = logging.getLogger(__name__)


def _bounds_to_float(lb, ub):
    if lb is None:
        lb = -math.inf
    if ub is None:
        ub = math.inf
    return lb, ub


def _print_var_set(var_set):
    s = f'{"LB":>12}{"UB":>12}    Var\n'

    for v in var_set:
        v_lb, v_ub = _bounds_to_float(*v.bounds)
        s += f'{v_lb:>12.2e}{v_ub:>12.2e}    {str(v)}\n'

    s += '\n'

    return s


def _check_var_bounds(m: BlockData, too_large: float):
    vars_without_bounds = ComponentSet()
    vars_with_large_bounds = ComponentSet()
    for v in m.component_data_objects(pyo.Var, descend_into=True):
        if v.is_fixed():
            continue
        if v.lb is None or v.ub is None:
            vars_without_bounds.add(v)
        elif v.lb <= -too_large or v.ub >= too_large:
            vars_with_large_bounds.add(v)

    return vars_without_bounds, vars_with_large_bounds


def _print_coefficients(comp_map):
    s = ''
    for c, der_bounds in comp_map.items():
        s += str(c)
        s += '\n'
        s += f'    {"Coef LB":>12}{"Coef UB":>12}    Var\n'
        for v, der_lb, der_ub in der_bounds:
            s += f'    {der_lb:>12.2e}{der_ub:>12.2e}    {str(v)}\n'
        s += '\n'
    return s


def _check_coefficients(
    comp, expr, too_large, too_small, largs_coef_map, small_coef_map
):
    ders = reverse_sd(expr)
    for _v, _der in ders.items():
        if getattr(_v, 'ctype', None) is Var:
            if _v.is_fixed():
                continue
            der_lb, der_ub = compute_bounds_on_expr(_der)
            der_lb, der_ub = _bounds_to_float(der_lb, der_ub)
            if der_lb <= -too_large or der_ub >= too_large:
                if comp not in largs_coef_map:
                    largs_coef_map[comp] = list()
                largs_coef_map[comp].append((_v, der_lb, der_ub))
            if abs(der_lb) <= too_small and abs(der_ub) < too_small:
                if der_lb != 0 or der_ub != 0:
                    if comp not in small_coef_map:
                        small_coef_map[comp] = list()
                    small_coef_map[comp].append((_v, der_lb, der_ub))


def report_scaling(
    m: BlockData, too_large: float = 5e4, too_small: float = 1e-6
) -> bool:
    """
    This function logs potentially poorly scaled parts of the model.
    It requires that all variables be bounded.

    It is important to note that this check is neither necessary nor sufficient
    to ensure a well-scaled model. However, it is a useful tool to help identify
    problematic parts of a model.

    This function uses symbolic differentiation and interval arithmetic
    to compute bounds on each entry in the jacobian of the constraints.

    Note that logging has to be turned on to get the output

    Parameters
    ----------
    m: BlockData
        The pyomo model or block
    too_large: float
        Values above too_large will generate a log entry
    too_small: float
        Coefficients below too_small will generate a log entry

    Returns
    -------
    success: bool
        Returns False if any potentially poorly scaled components were found
    """
    vars_without_bounds, vars_with_large_bounds = _check_var_bounds(m, too_large)

    cons_with_large_bounds = dict()
    cons_with_large_coefficients = dict()
    cons_with_small_coefficients = dict()

    objs_with_large_coefficients = pyo.ComponentMap()
    objs_with_small_coefficients = pyo.ComponentMap()

    for c in m.component_data_objects(pyo.Constraint, active=True, descend_into=True):
        _check_coefficients(
            c,
            c.body,
            too_large,
            too_small,
            cons_with_large_coefficients,
            cons_with_small_coefficients,
        )

    for c in m.component_data_objects(pyo.Constraint, active=True, descend_into=True):
        c_lb, c_ub = compute_bounds_on_expr(c.body)
        c_lb, c_ub = _bounds_to_float(c_lb, c_ub)
        if c_lb <= -too_large or c_ub >= too_large:
            cons_with_large_bounds[c] = (c_lb, c_ub)

    for c in m.component_data_objects(pyo.Objective, active=True, descend_into=True):
        _check_coefficients(
            c,
            c.expr,
            too_large,
            too_small,
            objs_with_large_coefficients,
            objs_with_small_coefficients,
        )

    s = '\n\n'

    if len(vars_without_bounds) > 0:
        s += 'The following variables are not bounded. Please add bounds.\n'
        s += _print_var_set(vars_without_bounds)

    if len(vars_with_large_bounds) > 0:
        s += 'The following variables have large bounds. Please scale them.\n'
        s += _print_var_set(vars_with_large_bounds)

    if len(objs_with_large_coefficients) > 0:
        s += 'The following objectives have potentially large coefficients. Please scale them.\n'
        s += _print_coefficients(objs_with_large_coefficients)

    if len(objs_with_small_coefficients) > 0:
        s += 'The following objectives have small coefficients.\n'
        s += _print_coefficients(objs_with_small_coefficients)

    if len(cons_with_large_coefficients) > 0:
        s += 'The following constraints have potentially large coefficients. Please scale them.\n'
        s += _print_coefficients(cons_with_large_coefficients)

    if len(cons_with_small_coefficients) > 0:
        s += 'The following constraints have small coefficients.\n'
        s += _print_coefficients(cons_with_small_coefficients)

    if len(cons_with_large_bounds) > 0:
        s += 'The following constraints have bodies with large bounds. Please scale them.\n'
        s += f'{"LB":>12}{"UB":>12}    Constraint\n'
        for c, (c_lb, c_ub) in cons_with_large_bounds.items():
            s += f'{c_lb:>12.2e}{c_ub:>12.2e}    {str(c)}\n'

    if (
        len(vars_without_bounds) > 0
        or len(vars_with_large_bounds) > 0
        or len(cons_with_large_coefficients) > 0
        or len(cons_with_small_coefficients) > 0
        or len(objs_with_small_coefficients) > 0
        or len(objs_with_large_coefficients) > 0
        or len(cons_with_large_bounds) > 0
    ):
        logger.info(s)
        return False
    return True



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import pyomo.common.unittest as unittest

from pyomo.common.env import CtypesEnviron


class TestCtypesEnviron(unittest.TestCase):
    def test_temp_env_str(self):
        orig_env = CtypesEnviron()
        orig_env_has_1 = 'TEST_ENV_1' in orig_env
        orig_env_has_2 = 'TEST_ENV_2' in orig_env

        # Ensure that TEST_ENV_1 is not in the environment
        if 'TEST_ENV_1' in orig_env:
            del orig_env['TEST_ENV_1']
        orig_env['TEST_ENV_2'] = "test value: 2"

        self.assertIsNone(os.environ.get('TEST_ENV_1', None))
        self.assertEqual(os.environ['TEST_ENV_2'], "test value: 2")
        for interface in orig_env.interfaces:
            self.assertIsNone(interface.dll.wgetenv(u'TEST_ENV_1'))
            self.assertIsNone(interface.dll.getenv(b'TEST_ENV_1'))
            self.assertEqual(interface.dll.wgetenv(u'TEST_ENV_2'), u"test value: 2")
            self.assertEqual(interface.dll.getenv(b'TEST_ENV_2'), b"test value: 2")

        with CtypesEnviron(TEST_ENV_1="test value: 1") as env:
            self.assertEqual(os.environ['TEST_ENV_1'], "test value: 1")
            self.assertEqual(os.environ['TEST_ENV_2'], "test value: 2")
            for interface in env.interfaces:
                self.assertEqual(interface.dll.wgetenv(u'TEST_ENV_1'), u"test value: 1")
                self.assertEqual(interface.dll.getenv(b'TEST_ENV_1'), b"test value: 1")
                self.assertEqual(interface.dll.wgetenv(u'TEST_ENV_2'), u"test value: 2")
                self.assertEqual(interface.dll.getenv(b'TEST_ENV_2'), b"test value: 2")

            del env['TEST_ENV_2']
            self.assertIsNone(os.environ.get('TEST_ENV_2', None))
            for interface in env.interfaces:
                self.assertIsNone(interface.dll.wgetenv(u'TEST_ENV_2'))
                self.assertIsNone(interface.dll.getenv(b'TEST_ENV_2'))

        self.assertIsNone(os.environ.get('TEST_ENV_1', None))
        self.assertEqual(os.environ['TEST_ENV_2'], "test value: 2")
        for interface in orig_env.interfaces:
            self.assertIsNone(interface.dll.wgetenv(u'TEST_ENV_1'))
            self.assertIsNone(interface.dll.getenv(b'TEST_ENV_1'))
            self.assertEqual(interface.dll.wgetenv(u'TEST_ENV_2'), u"test value: 2")
            self.assertEqual(interface.dll.getenv(b'TEST_ENV_2'), b"test value: 2")

        orig_env.restore()
        self.assertEqual(orig_env_has_1, 'TEST_ENV_1' in os.environ)
        self.assertEqual(orig_env_has_2, 'TEST_ENV_2' in os.environ)

    def test_temp_env_unicode(self):
        orig_env = CtypesEnviron()
        orig_env_has_1 = u'TEST_ENV_1' in orig_env
        orig_env_has_2 = u'TEST_ENV_2' in orig_env

        # Ensure that TEST_ENV_1 is not in the environment
        if u'TEST_ENV_1' in orig_env:
            del orig_env[u'TEST_ENV_1']
        orig_env[u'TEST_ENV_2'] = u"test value: 2"

        self.assertIsNone(os.environ.get(u'TEST_ENV_1', None))
        self.assertEqual(os.environ[u'TEST_ENV_2'], "test value: 2")
        for interface in orig_env.interfaces:
            self.assertIsNone(interface.dll.wgetenv(u'TEST_ENV_1'))
            self.assertIsNone(interface.dll.getenv(b'TEST_ENV_1'))
            self.assertEqual(interface.dll.wgetenv(u'TEST_ENV_2'), u"test value: 2")
            self.assertEqual(interface.dll.getenv(b'TEST_ENV_2'), b"test value: 2")

        with CtypesEnviron(TEST_ENV_1=u"test value: 1") as env:
            self.assertEqual(os.environ[u'TEST_ENV_1'], u"test value: 1")
            self.assertEqual(os.environ[u'TEST_ENV_2'], u"test value: 2")
            for interface in env.interfaces:
                self.assertEqual(interface.dll.wgetenv(u'TEST_ENV_1'), u"test value: 1")
                self.assertEqual(interface.dll.getenv(b'TEST_ENV_1'), b"test value: 1")
                self.assertEqual(interface.dll.wgetenv(u'TEST_ENV_2'), u"test value: 2")
                self.assertEqual(interface.dll.getenv(b'TEST_ENV_2'), b"test value: 2")

            del env[u'TEST_ENV_2']
            self.assertIsNone(os.environ.get(u'TEST_ENV_2', None))
            for interface in env.interfaces:
                self.assertIsNone(interface.dll.wgetenv(u'TEST_ENV_2'))
                self.assertIsNone(interface.dll.getenv(b'TEST_ENV_2'))

        self.assertIsNone(os.environ.get(u'TEST_ENV_1', None))
        self.assertEqual(os.environ[u'TEST_ENV_2'], u"test value: 2")
        for interface in orig_env.interfaces:
            self.assertIsNone(interface.dll.wgetenv(u'TEST_ENV_1'))
            self.assertIsNone(interface.dll.getenv(b'TEST_ENV_1'))
            self.assertEqual(interface.dll.wgetenv(u'TEST_ENV_2'), u"test value: 2")
            self.assertEqual(interface.dll.getenv(b'TEST_ENV_2'), b"test value: 2")

        orig_env.restore()
        self.assertEqual(orig_env_has_1, u'TEST_ENV_1' in os.environ)
        self.assertEqual(orig_env_has_2, u'TEST_ENV_2' in os.environ)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"This is the deprecated pyomo.core.base.template_expr module"

from pyomo.core.expr.template_expr import (
    IndexTemplate,
    _GetItemIndexer,
    TemplateExpressionError,
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections.abc
import pickle

import pyomo.common.unittest as unittest
from pyomo.kernel import pprint
from pyomo.core.tests.unit.kernel.test_dict_container import (
    _TestActiveDictContainerBase,
)
from pyomo.core.kernel.base import ICategorizedObject
from pyomo.core.kernel.suffix import (
    ISuffix,
    suffix,
    suffix_dict,
    export_suffix_generator,
    import_suffix_generator,
    local_suffix_generator,
    suffix_generator,
)
from pyomo.core.kernel.variable import variable
from pyomo.core.kernel.constraint import constraint, constraint_list
from pyomo.core.kernel.block import block, block_dict


class Test_suffix(unittest.TestCase):
    def test_pprint(self):
        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        v = variable()
        clist = constraint_list([constraint()])
        s = suffix()
        s[v] = 1
        s[clist] = None
        pprint(s)
        b = block()
        b.s = s
        pprint(s)
        pprint(b)
        m = block()
        m.b = b
        pprint(s)
        pprint(b)
        pprint(m)

        pprint({'a': 1, 'b': 2})

    def test_str(self):
        s = suffix()
        self.assertEqual(str(s), "<suffix>")
        b = block()
        b.s = s
        self.assertEqual(str(s), "s")

    def test_ctype(self):
        s = suffix()
        self.assertIs(s.ctype, ISuffix)
        self.assertIs(type(s), suffix)
        self.assertIs(type(s)._ctype, ISuffix)

    def test_pickle(self):
        s = suffix(direction=suffix.EXPORT, datatype=suffix.FLOAT)
        self.assertEqual(s.direction, suffix.EXPORT)
        self.assertEqual(s.datatype, suffix.FLOAT)
        self.assertEqual(s.parent, None)
        sup = pickle.loads(pickle.dumps(s))
        self.assertEqual(sup.direction, suffix.EXPORT)
        self.assertEqual(sup.datatype, suffix.FLOAT)
        self.assertEqual(sup.parent, None)
        b = block()
        b.s = s
        self.assertIs(s.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        sup = bup.s
        self.assertEqual(sup.direction, suffix.EXPORT)
        self.assertEqual(sup.datatype, suffix.FLOAT)
        self.assertIs(sup.parent, bup)
        b.v = variable(lb=1)
        b.s[b.v] = 1.0
        bup = pickle.loads(pickle.dumps(b))
        sup = bup.s
        vup = bup.v
        self.assertEqual(sup[vup], 1.0)

    def test_init(self):
        s = suffix()
        self.assertTrue(s.parent is None)
        self.assertEqual(s.ctype, ISuffix)
        self.assertEqual(s.direction, suffix.LOCAL)
        self.assertEqual(s.datatype, suffix.FLOAT)

    def test_type(self):
        s = suffix()
        self.assertTrue(isinstance(s, ICategorizedObject))
        self.assertTrue(isinstance(s, collections.abc.Mapping))
        self.assertTrue(isinstance(s, collections.abc.MutableMapping))
        self.assertTrue(issubclass(type(s), collections.abc.Mapping))
        self.assertTrue(issubclass(type(s), collections.abc.MutableMapping))

    def test_import_export_enabled(self):
        s = suffix()
        s.direction = suffix.LOCAL
        self.assertEqual(s.direction, suffix.LOCAL)
        self.assertEqual(s.export_enabled(), False)
        self.assertEqual(s.import_enabled(), False)
        s.direction = suffix.IMPORT
        self.assertEqual(s.direction, suffix.IMPORT)
        self.assertEqual(s.export_enabled(), False)
        self.assertEqual(s.import_enabled(), True)
        s.direction = suffix.EXPORT
        self.assertEqual(s.direction, suffix.EXPORT)
        self.assertEqual(s.export_enabled(), True)
        self.assertEqual(s.import_enabled(), False)
        s.direction = suffix.IMPORT_EXPORT
        self.assertEqual(s.direction, suffix.IMPORT_EXPORT)
        self.assertEqual(s.export_enabled(), True)
        self.assertEqual(s.import_enabled(), True)
        with self.assertRaises(ValueError):
            s.direction = 'export'

    def test_datatype(self):
        s = suffix()
        s.datatype = suffix.FLOAT
        self.assertEqual(s.datatype, suffix.FLOAT)
        s.datatype = suffix.INT
        self.assertEqual(s.datatype, suffix.INT)
        with self.assertRaises(ValueError):
            s.datatype = 'something'

    def test_clear(self):
        x = variable()
        y = variable()

        s = suffix()
        s[x] = 1.0
        s[y] = None

        self.assertEqual(len(s), 2)
        s.clear()
        self.assertEqual(len(s), 0)

    def test_del(self):
        x = variable()

        s = suffix()
        s[x] = 1.0

        self.assertEqual(len(s), 1)
        del s[x]
        self.assertEqual(len(s), 0)
        with self.assertRaises(KeyError):
            del s[x]

    def test_name(self):
        s = suffix()
        self.assertTrue(s.parent is None)
        self.assertEqual(s.local_name, None)
        self.assertEqual(s.name, None)

        model = block()
        model.s = s
        self.assertTrue(model.parent is None)
        self.assertTrue(s.parent is model)
        self.assertEqual(s.local_name, "s")
        self.assertEqual(s.name, "s")

        b = block()
        b.model = model
        self.assertTrue(b.parent is None)
        self.assertTrue(model.parent is b)
        self.assertTrue(s.parent is model)
        self.assertEqual(s.local_name, "s")
        self.assertEqual(s.name, "model.s")

        bdict = block_dict()
        bdict[0] = b
        self.assertTrue(bdict.parent is None)
        self.assertTrue(b.parent is bdict)
        self.assertTrue(model.parent is b)
        self.assertTrue(s.parent is model)
        self.assertEqual(s.local_name, "s")
        self.assertEqual(s.name, "[0].model.s")

        m = block()
        m.bdict = bdict
        self.assertTrue(m.parent is None)
        self.assertTrue(bdict.parent is m)
        self.assertTrue(b.parent is bdict)
        self.assertTrue(model.parent is b)
        self.assertTrue(s.parent is model)
        self.assertEqual(s.local_name, "s")
        self.assertEqual(s.name, "bdict[0].model.s")

    def test_active(self):
        s = suffix()
        with self.assertRaises(AttributeError):
            s.active = False

        model = block()
        model.s = s
        b = block()
        b.model = model
        bdict = block_dict()
        bdict[0] = b
        bdict[None] = block()
        m = block()
        m.bdict = bdict

        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(s.active, True)

        m.deactivate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(bdict.active, False)
        self.assertEqual(bdict[None].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(s.active, False)

        m.activate(shallow=False)

        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(s.active, True)

        m.deactivate()

        self.assertEqual(m.active, False)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(s.active, True)

        m.deactivate(shallow=False)

        self.assertEqual(m.active, False)
        self.assertEqual(bdict.active, False)
        self.assertEqual(bdict[None].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(s.active, False)

        m.activate()

        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, False)
        self.assertEqual(bdict[None].active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(model.active, False)
        self.assertEqual(s.active, False)

        m.activate(shallow=False)

        self.assertEqual(m.active, True)
        self.assertEqual(bdict.active, True)
        self.assertEqual(bdict[None].active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(model.active, True)
        self.assertEqual(s.active, True)

    def test_export_suffix_generator(self):
        m = block()
        m.s0 = suffix(direction=suffix.LOCAL)
        m.s0i = suffix(direction=suffix.LOCAL, datatype=suffix.INT)
        m.s1 = suffix(direction=suffix.IMPORT_EXPORT)
        m.s1i = suffix(direction=suffix.IMPORT_EXPORT, datatype=suffix.INT)
        m.s2 = suffix(direction=suffix.IMPORT)
        m.s2i = suffix(direction=suffix.IMPORT, datatype=suffix.INT)
        m.s3 = suffix(direction=suffix.EXPORT)
        m.s3i = suffix(direction=suffix.EXPORT, datatype=suffix.INT)
        m.b = block()
        m.b.s0 = suffix(direction=suffix.LOCAL)
        m.b.s0i = suffix(direction=suffix.LOCAL, datatype=suffix.INT)
        m.b.s1 = suffix(direction=suffix.IMPORT_EXPORT)
        m.b.s1i = suffix(direction=suffix.IMPORT_EXPORT, datatype=suffix.INT)
        m.b.s2 = suffix(direction=suffix.IMPORT)
        m.b.s2i = suffix(direction=suffix.IMPORT, datatype=suffix.INT)
        m.b.s3 = suffix(direction=suffix.EXPORT)
        m.b.s3i = suffix(direction=suffix.EXPORT, datatype=suffix.INT)
        # default
        self.assertEqual(
            [id(c_) for c_ in export_suffix_generator(m)],
            [
                id(m.s1),
                id(m.s1i),
                id(m.s3),
                id(m.s3i),
                id(m.b.s1),
                id(m.b.s1i),
                id(m.b.s3),
                id(m.b.s3i),
            ],
        )
        # descend_into=False
        self.assertEqual(
            [id(c_) for c_ in export_suffix_generator(m, descend_into=False)],
            [id(m.s1), id(m.s1i), id(m.s3), id(m.s3i)],
        )
        # datatype=INT
        self.assertEqual(
            [id(c_) for c_ in export_suffix_generator(m, datatype=suffix.INT)],
            [id(m.s1i), id(m.s3i), id(m.b.s1i), id(m.b.s3i)],
        )
        # active=True
        m.s1.deactivate()
        m.b.deactivate()
        self.assertEqual(
            [id(c_) for c_ in export_suffix_generator(m, active=True)],
            [id(m.s1i), id(m.s3), id(m.s3i)],
        )

    def test_import_suffix_generator(self):
        m = block()
        m.s0 = suffix(direction=suffix.LOCAL)
        m.s0i = suffix(direction=suffix.LOCAL, datatype=suffix.INT)
        m.s1 = suffix(direction=suffix.IMPORT_EXPORT)
        m.s1i = suffix(direction=suffix.IMPORT_EXPORT, datatype=suffix.INT)
        m.s2 = suffix(direction=suffix.IMPORT)
        m.s2i = suffix(direction=suffix.IMPORT, datatype=suffix.INT)
        m.s3 = suffix(direction=suffix.EXPORT)
        m.s3i = suffix(direction=suffix.EXPORT, datatype=suffix.INT)
        m.b = block()
        m.b.s0 = suffix(direction=suffix.LOCAL)
        m.b.s0i = suffix(direction=suffix.LOCAL, datatype=suffix.INT)
        m.b.s1 = suffix(direction=suffix.IMPORT_EXPORT)
        m.b.s1i = suffix(direction=suffix.IMPORT_EXPORT, datatype=suffix.INT)
        m.b.s2 = suffix(direction=suffix.IMPORT)
        m.b.s2i = suffix(direction=suffix.IMPORT, datatype=suffix.INT)
        m.b.s3 = suffix(direction=suffix.EXPORT)
        m.b.s3i = suffix(direction=suffix.EXPORT, datatype=suffix.INT)
        # default
        self.assertEqual(
            [id(c_) for c_ in import_suffix_generator(m)],
            [
                id(m.s1),
                id(m.s1i),
                id(m.s2),
                id(m.s2i),
                id(m.b.s1),
                id(m.b.s1i),
                id(m.b.s2),
                id(m.b.s2i),
            ],
        )
        # descend_into=False
        self.assertEqual(
            [id(c_) for c_ in import_suffix_generator(m, descend_into=False)],
            [id(m.s1), id(m.s1i), id(m.s2), id(m.s2i)],
        )
        # datatype=INT
        self.assertEqual(
            [id(c_) for c_ in import_suffix_generator(m, datatype=suffix.INT)],
            [id(m.s1i), id(m.s2i), id(m.b.s1i), id(m.b.s2i)],
        )
        # active=True
        m.s1.deactivate()
        m.b.deactivate()
        self.assertEqual(
            [id(c_) for c_ in import_suffix_generator(m, active=True)],
            [id(m.s1i), id(m.s2), id(m.s2i)],
        )

    def test_local_suffix_generator(self):
        m = block()
        m.s0 = suffix(direction=suffix.LOCAL)
        m.s0i = suffix(direction=suffix.LOCAL, datatype=suffix.INT)
        m.s1 = suffix(direction=suffix.IMPORT_EXPORT)
        m.s1i = suffix(direction=suffix.IMPORT_EXPORT, datatype=suffix.INT)
        m.s2 = suffix(direction=suffix.IMPORT)
        m.s2i = suffix(direction=suffix.IMPORT, datatype=suffix.INT)
        m.s3 = suffix(direction=suffix.EXPORT)
        m.s3i = suffix(direction=suffix.EXPORT, datatype=suffix.INT)
        m.b = block()
        m.b.s0 = suffix(direction=suffix.LOCAL)
        m.b.s0i = suffix(direction=suffix.LOCAL, datatype=suffix.INT)
        m.b.s1 = suffix(direction=suffix.IMPORT_EXPORT)
        m.b.s1i = suffix(direction=suffix.IMPORT_EXPORT, datatype=suffix.INT)
        m.b.s2 = suffix(direction=suffix.IMPORT)
        m.b.s2i = suffix(direction=suffix.IMPORT, datatype=suffix.INT)
        m.b.s3 = suffix(direction=suffix.EXPORT)
        m.b.s3i = suffix(direction=suffix.EXPORT, datatype=suffix.INT)
        # default
        self.assertEqual(
            [id(c_) for c_ in local_suffix_generator(m)],
            [id(m.s0), id(m.s0i), id(m.b.s0), id(m.b.s0i)],
        )
        # descend_into=False
        self.assertEqual(
            [id(c_) for c_ in local_suffix_generator(m, descend_into=False)],
            [id(m.s0), id(m.s0i)],
        )
        # datatype=INT
        self.assertEqual(
            [id(c_) for c_ in local_suffix_generator(m, datatype=suffix.INT)],
            [id(m.s0i), id(m.b.s0i)],
        )
        # active=True
        m.s0.deactivate()
        m.b.deactivate()
        self.assertEqual(
            [id(c_) for c_ in local_suffix_generator(m, active=True)], [id(m.s0i)]
        )

    def test_suffix_generator(self):
        m = block()
        m.s0 = suffix(direction=suffix.LOCAL)
        m.s0i = suffix(direction=suffix.LOCAL, datatype=suffix.INT)
        m.s1 = suffix(direction=suffix.IMPORT_EXPORT)
        m.s1i = suffix(direction=suffix.IMPORT_EXPORT, datatype=suffix.INT)
        m.s2 = suffix(direction=suffix.IMPORT)
        m.s2i = suffix(direction=suffix.IMPORT, datatype=suffix.INT)
        m.s3 = suffix(direction=suffix.EXPORT)
        m.s3i = suffix(direction=suffix.EXPORT, datatype=suffix.INT)
        m.b = block()
        m.b.s0 = suffix(direction=suffix.LOCAL)
        m.b.s0i = suffix(direction=suffix.LOCAL, datatype=suffix.INT)
        m.b.s1 = suffix(direction=suffix.IMPORT_EXPORT)
        m.b.s1i = suffix(direction=suffix.IMPORT_EXPORT, datatype=suffix.INT)
        m.b.s2 = suffix(direction=suffix.IMPORT)
        m.b.s2i = suffix(direction=suffix.IMPORT, datatype=suffix.INT)
        m.b.s3 = suffix(direction=suffix.EXPORT)
        m.b.s3i = suffix(direction=suffix.EXPORT, datatype=suffix.INT)
        # default
        self.assertEqual(
            [id(c_) for c_ in suffix_generator(m)],
            [
                id(m.s0),
                id(m.s0i),
                id(m.s1),
                id(m.s1i),
                id(m.s2),
                id(m.s2i),
                id(m.s3),
                id(m.s3i),
                id(m.b.s0),
                id(m.b.s0i),
                id(m.b.s1),
                id(m.b.s1i),
                id(m.b.s2),
                id(m.b.s2i),
                id(m.b.s3),
                id(m.b.s3i),
            ],
        )
        # descend_into=False
        self.assertEqual(
            [id(c_) for c_ in suffix_generator(m, descend_into=False)],
            [
                id(m.s0),
                id(m.s0i),
                id(m.s1),
                id(m.s1i),
                id(m.s2),
                id(m.s2i),
                id(m.s3),
                id(m.s3i),
            ],
        )
        # datatype=INT
        self.assertEqual(
            [id(c_) for c_ in suffix_generator(m, datatype=suffix.INT)],
            [
                id(m.s0i),
                id(m.s1i),
                id(m.s2i),
                id(m.s3i),
                id(m.b.s0i),
                id(m.b.s1i),
                id(m.b.s2i),
                id(m.b.s3i),
            ],
        )
        # active=True
        m.s1.deactivate()
        m.b.deactivate()
        self.assertEqual(
            [id(c_) for c_ in suffix_generator(m, active=True)],
            [id(m.s0), id(m.s0i), id(m.s1i), id(m.s2), id(m.s2i), id(m.s3), id(m.s3i)],
        )

    #
    # These methods are deprecated
    #

    def test_set_all_values(self):
        x = variable()
        y = variable()

        s = suffix()
        s[x] = 1.0
        s[y] = None

        self.assertEqual(s[x], 1.0)
        self.assertEqual(s[y], None)
        s.set_all_values(0)
        self.assertEqual(s[x], 0)
        self.assertEqual(s[y], 0)

    def test_clear_all_values(self):
        x = variable()
        y = variable()

        s = suffix()
        s[x] = 1.0
        s[y] = None

        self.assertEqual(len(s), 2)
        s.clear_all_values()
        self.assertEqual(len(s), 0)
        s.clear_all_values()

    def test_clear_value(self):
        x = variable()

        s = suffix()
        s[x] = 1.0

        self.assertEqual(len(s), 1)
        s.clear_value(x)
        self.assertEqual(len(s), 0)
        s.clear_value(x)

    def test_getset_direction(self):
        s = suffix()
        s.set_direction(suffix.LOCAL)
        self.assertEqual(s.get_direction(), suffix.LOCAL)
        s.set_direction(suffix.IMPORT)
        self.assertEqual(s.get_direction(), suffix.IMPORT)
        s.set_direction(suffix.EXPORT)
        self.assertEqual(s.get_direction(), suffix.EXPORT)
        s.set_direction(suffix.IMPORT_EXPORT)
        self.assertEqual(s.get_direction(), suffix.IMPORT_EXPORT)
        with self.assertRaises(ValueError):
            s.set_direction('export')

    def test_getset_datatype(self):
        s = suffix()
        s.set_datatype(suffix.FLOAT)
        self.assertEqual(s.get_datatype(), suffix.FLOAT)
        s.set_datatype(suffix.INT)
        self.assertEqual(s.get_datatype(), suffix.INT)
        with self.assertRaises(ValueError):
            s.set_datatype('something')


class Test_suffix_dict(_TestActiveDictContainerBase, unittest.TestCase):
    _container_type = suffix_dict
    _ctype_factory = lambda self: suffix()


if __name__ == "__main__":
    unittest.main()
